import os
import json

from medcat.cdb.cdb import CDB
from medcat.tokenizing.tokenizers import BaseTokenizer
from medcat.components.ner.trf.transformers_ner import TransformersNER
from medcat.config.config_transformers_ner import ConfigTransformersNER
from medcat.utils.legacy.convert_cdb import get_cdb_from_old


exp_cat_config_name = "cat_config.json"
exp_trf_ner_config_name = "config.json"
exp_cdb_name = "cdb.dat"


def get_cnf(cnf_path: str) -> ConfigTransformersNER:
    cnf = ConfigTransformersNER()
    with open(cnf_path) as f:
        old_stuff = json.load(f)
    cnf.merge_config(old_stuff)
    return cnf


def get_trf_ner_from_old(old_path: str, tokenizer: BaseTokenizer
                         ) -> TransformersNER:
    config = get_cnf(os.path.join(old_path, exp_cat_config_name))
    config.general.model_name = old_path
    cdb: CDB = get_cdb_from_old(os.path.join(old_path, exp_cdb_name))
    trf_ner = TransformersNER.create_new(cdb, tokenizer, config)
    trf_ner._component.create_eval_pipeline()
    return trf_ner



import os
from typing import (Optional, Iterator, Iterable, TypeVar, cast, Type, Any,
                    Literal)
from typing import Protocol, runtime_checkable
from typing_extensions import Self
import logging
from datetime import datetime
from contextlib import contextmanager

from pydantic import BaseModel, Field, ValidationError

from medcat import __version__ as medcat_version
from medcat.utils.defaults import workers
from medcat.utils.envsnapshot import Environment, get_environment_info
from medcat.utils.iterutils import callback_iterator
from medcat.utils.defaults import (
    avoid_legacy_conversion, doing_legacy_conversion_message,
    LegacyConversionDisabledError)
from medcat.storage.serialisables import SerialisingStrategy
from medcat.storage.serialisers import deserialise


logger = logging.getLogger(__name__)


class SerialisableBaseModel(BaseModel):
    """The base serialisable config."""

    def get_strategy(self) -> SerialisingStrategy:
        return SerialisingStrategy.SERIALISABLES_AND_DICT

    @classmethod
    def get_init_attrs(cls) -> list[str]:
        return []

    @classmethod
    def ignore_attrs(cls) -> list[str]:
        return []

    @classmethod
    def include_properties(cls) -> list[str]:
        return []

    def merge_config(self, other: dict):
        """Merge this config with another config's (partial) model dump.

        The exepctation is that the `other` dict is a partial model dump.
        Values specified there are overwritten into the current config.
        Values not specified there are left intact.

        The `other` config can have keys/values that do not exist in the
        config or sub-config. And they will be added where possible.

        Args:
            other (dict): The model dump

        Raises:
            IncorrectConfigValues: If unable to set the attribute,
                trying to set incorrect value, or trying to set sub-config
                values in an incorrect format (non-dict).
        """
        for k, v in other.items():
            if not hasattr(self, k):
                try:
                    setattr(self, k, v)
                except (ValidationError, ValueError) as e:
                    raise IncorrectConfigValues(
                        type(self), k, type(None), v
                    ) from e
                continue
            cur_v = getattr(self, k)
            if isinstance(cur_v, SerialisableBaseModel):
                if not isinstance(v, dict):
                    raise IncorrectConfigValues(
                        type(self), k, type(cur_v), v)
                cur_v.merge_config(v)
            else:
                try:
                    setattr(self, k, v)
                except ValidationError as e:
                    raise IncorrectConfigValues(
                        type(self), k, type(cur_v), v
                    ) from e

    @classmethod
    def load(cls, path: str) -> Self:
        if os.path.isfile(path) and path.endswith(".dat"):
            if avoid_legacy_conversion():
                raise LegacyConversionDisabledError(cls.__name__)
            doing_legacy_conversion_message(logger, cls.__name__, path)
            from medcat.utils.legacy.convert_config import (
                get_config_from_old_per_cls)
            return cast(Self, get_config_from_old_per_cls(path, cls))
        obj = deserialise(path)
        if not isinstance(obj, cls):
            raise ValueError(f"The path '{path}' is not a {cls.__name__}!")
        return obj


class IncorrectConfigValues(ValueError):

    def __init__(self, cls: Type, attr_name: str,
                 exp_type: Type, got: Any):
        super().__init__(f"Incorrect attribute set for {cls}.{attr_name}. "
                         f"Expected {exp_type}, but got {type(got)}: {got}")


@runtime_checkable
class PotentiallyDirty(Protocol):
    @property
    def is_dirty(self) -> bool:
        pass

    def mark_clean(self) -> None:
        pass


class DirtiableBaseModel(SerialisableBaseModel):
    _is_dirty: bool = False

    def __setattr__(self, name: str, value: Any):
        if name != '_is_dirty' and name in self.__annotations__:
            current = getattr(self, name, None)
            if current != value:
                self._is_dirty = True
        super().__setattr__(name, value)

    @property
    def is_dirty(self) -> bool:
        if self._is_dirty:
            return True
        for pn, part in self.__dict__.items():
            if isinstance(part, PotentiallyDirty):
                if part.is_dirty:
                    return True
        return False

    def mark_clean(self):
        self._is_dirty = False
        for part in self.__dict__.values():
            if isinstance(part, PotentiallyDirty):
                part.mark_clean()


class ComponentConfig(DirtiableBaseModel):
    comp_name: str = 'default'
    """The name of the component.

    If a custom implementation is required, it needs to be registered
    using `medcat.components.types.register_core_component(
            <core component type>, <component name>, <implementing class>)
    By default, only the 'default' component is registered.
    """


class NLPConfig(SerialisableBaseModel):
    disabled_components: list = ['ner', 'parser', 'vectors', 'textcat',
                                 'entity_linker', 'sentencizer',
                                 'entity_ruler', 'merge_noun_chunks',
                                 'merge_entities', 'merge_subtokens']
    """The list of components that will be disabled for the NLP.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    provider: str = 'regex'
    """The NLP provider.

    Currently only regex and spacy are natively supported.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    modelname: str = 'en_core_web_md'
    """What model will be used for tokenization.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """

    # NOTE: this will allow for more config entries
    #       since we don't know what other implementations may require

    class Config:
        extra = 'allow'
        validate_assignment = True


class UsageMonitor(SerialisableBaseModel):
    enabled: Literal[True, False, 'auto'] = False
    r"""Whether usage monitoring is enabled (True), disabled (False),
    or automatic ('auto').
    If set to False, no logging is performed.
    If set to True, logs are saved in the location specified by `log_folder`.
    If set to 'auto', logs will be automatically enabled or disabled based on
    environmenta variable (`MEDCAT_LOGS` - setting it to False or 0
    disabled logging) and distributed according to the OS preferred logs
    location (`MEDCAT_LOGS_LOCATION`).
    The defaults for the location are:
     - For Linux: ~/.local/share/medcat/logs/
     - For Windows: C:\Users\%USERNAME%\.cache\medcat\logs\
    """
    batch_size: int = 100
    """Number of logged events to write at once."""
    file_prefix: str = "usage_"
    """The prefix for logged files. The suffix will be the model hash."""
    log_folder: str = "."
    """The folder which contains the usage logs. In certain situations,
    it may make sense to keep this separate from the overall logs.
    NOTE: Does not take affect if `enabled` is set to 'auto'"""


class General(SerialisableBaseModel):
    """The general part of the config"""
    nlp: NLPConfig = NLPConfig()
    # checkpoint: CheckPoint = CheckPoint()
    usage_monitor: UsageMonitor = UsageMonitor()
    """Checkpointing config"""
    log_level: int = logging.INFO
    """Logging config for everything | 'tagger' can be disabled,
    but will cause a drop in performance"""
    log_format: str = '%(levelname)s:%(name)s: %(message)s'
    log_path: str = './medcat.log'
    separator: str = '~'
    """Separator that will be used to merge tokens of a name.
    Once a CDB is built this should always stay the same."""
    spell_check: bool = True
    """Should we check spelling - note that this makes things much slower,
    use only if necessary. The only thing necessary for the spell checker
    to work is vocab.dat and cdb.dat built with concepts in the respective
    language."""
    diacritics: bool = False
    """Should we process diacritics - for languages other than English,
    symbols such as 'é, ë, ö' can be relevant. Note that this makes
    spell_check slower."""
    spell_check_deep: bool = False
    """If True the spell checker will try harder to find mistakes,
    this can slow down things drastically."""
    spell_check_len_limit: int = 7
    """Spelling will not be checked for words with length less than this"""
    show_nested_entities: bool = False
    """If set to True functions like get_entities and get_json will return
    nested_entities and overlaps"""
    full_unlink: bool = False
    """When unlinking a name from a concept should we do full_unlink
    (means unlink a name from all concepts, not just the one in question)"""
    workers: int = workers()
    """Number of workers used by a parallelizable pipeline component"""
    make_pretty_labels: Optional[str] = None
    """Should the labels of entities (shown in displacy) be pretty
    or just 'concept'. Slows down the annotation pipeline
    should not be used when annotating millions of documents.
    If `None` it will be the string "concept", if `short` it will be CUI,
    if `long` it will be CUI | Name | Confidence"""
    map_cui_to_group: bool = False
    """If the cdb.addl_info['cui2group'] is provided and this option enabled,
    each CUI will be mapped to the group"""
    simple_hash: bool = False
    """Whether to use a simple hash.

    NOTE: While using a simple hash is faster at save time, it is less
    reliable due to not taking into account all the details of the changes."""

    class Config:
        extra = 'allow'


class LinkingFilters(SerialisableBaseModel):
    """These describe the linking filters used alongside the model.

    When no CUIs nor excluded CUIs are specified (the sets are empty),
    all CUIs are accepted.
    If there are CUIs specified then only those will be accepted.
    If there are excluded CUIs specified, they are excluded.

    In some cases, there are extra filters as well as MedCATtrainer (MCT)
    export filters. These are expected to follow the following:
    extra_cui_filter ⊆ MCT filter ⊆ Model/config filter

    While any other CUIs can be included in the the extra CUI filter or
    the MCT filter, they would not have any real effect.
    """
    cuis: set[str] = set()
    cuis_exclude: set[str] = set()

    def __init__(self, **data):
        if 'cuis' in data:
            cuis = data['cuis']
            if isinstance(cuis, dict) and len(cuis) == 0:
                logger.warning("Loading an old model where "
                               "config.linking.filters.cuis has been "
                               "dict to an empty dict instead of an empty "
                               "set. Converting the dict to a set in memory "
                               "as that is what is expected. Please consider "
                               "saving the model again.")
                data['cuis'] = set(cuis.keys())
        super().__init__(**data)

    def check_filters(self, cui: str) -> bool:
        """Checks is a CUI in the filters

        Args:
            cui (str): The CUI in question

        Returns:
            bool: True if the CUI is allowed
        """
        if cui in self.cuis or not self.cuis:
            return cui not in self.cuis_exclude
        else:
            return False


class Linking(ComponentConfig):
    """The linking part of the config"""
    optim: dict = {'type': 'linear', 'base_lr': 1, 'min_lr': 0.00005}
    """Linear anneal"""
    # optim: dict = {'type': 'standard', 'lr': 1}
    context_vector_sizes: dict = {'xlong': 27, 'long': 18,
                                  'medium': 9, 'short': 3}
    """Context vector sizes that will be calculated and used for linking"""
    context_vector_weights: dict = {'xlong': 0.1, 'long': 0.4,
                                    'medium': 0.4, 'short': 0.1}
    """Weight of each vector in the similarity score - make trainable at
    some point. Should add up to 1."""
    filters: LinkingFilters = LinkingFilters()
    """Filters"""
    train: bool = True
    """Should it train or not, this is set automatically ignore in 99% of
    cases and do not set manually"""
    random_replacement_unsupervised: float = 0.80
    """If <1 during unsupervised training the detected term will be randomly
    replaced with a probability of 1 - random_replacement_unsupervised
    Replaced with a synonym used for that term"""
    disamb_length_limit: int = 3
    """All concepts below this will always be disambiguated"""
    filter_before_disamb: bool = False
    """If True it will filter before doing disamb. Useful for the trainer."""
    train_count_threshold: int = 1
    """Concepts that have seen less training examples than this will not be
    used for similarity calculation and will have a similarity of -1."""
    always_calculate_similarity: bool = False
    """Do we want to calculate context similarity even for concepts that are
    not ambiguous."""
    calculate_dynamic_threshold: bool = False
    """Concepts below this similarity will be ignored. Type can be
    static/dynamic - if dynamic each CUI has a different TH
    and it is calculated as the average confidence for that
    CUI * similarity_threshold. Take care that dynamic works only
    if the cdb was trained with calculate_dynamic_threshold = True."""
    similarity_threshold_type: str = 'static'
    similarity_threshold: float = 0.25
    negative_probability: float = 0.5
    """Probability for the negative context to be added for each
    positive addition"""
    negative_ignore_punct_and_num: bool = True
    """Do we ignore punct/num when negative sampling"""
    prefer_primary_name: float = 0.35
    """If >0 concepts for which a detection is its primary name
    will be preferred by that amount (0 to 1)"""
    prefer_frequent_concepts: float = 0.35
    """If >0 concepts that are more frequent will be preferred
    by a multiply of this amount"""
    subsample_after: int = 30000
    """DISABLED in code permanetly: Subsample during unsupervised
    training if a concept has received more than"""
    devalue_linked_concepts: bool = False
    """When adding a positive example, should it also be treated as Negative
    for concepts which link to the positive one via names (ambiguous names)."""
    context_ignore_center_tokens: bool = False
    """If true when the context of a concept is calculated (embedding)
    the words making that concept are not taken into account"""
    additional: Optional[Any] = None
    """Some additional config for non-default linkers.
    E.g the 2-step linker uses this for alpha calculations
    and learning rate for type contexts.
    """

    class Config:
        extra = 'allow'


class Preprocessing(SerialisableBaseModel):
    """The preprocessing part of the config"""
    words_to_skip: set = {'nos'}
    """This words will be completely ignored from concepts and from the text
    (must be a Set)"""
    keep_punct: set = {'.', ':'}
    """All punct will be skipped by default, here you can set what
    will be kept"""
    do_not_normalize: set[str] = {'VBD', 'VBG', 'VBN', 'VBP', 'JJS', 'JJR'}
    """Should specific word types be normalized: e.g. running -> run
    Values are detailed part-of-speech tags. See:
    - https://spacy.io/usage/linguistic-features#pos-tagging
    - Label scheme section per model at https://spacy.io/models/en"""
    skip_stopwords: bool = False
    """Should stopwords be skipped/ignored when processing input"""
    min_len_normalize: int = 5
    """Nothing below this length will ever be normalized (input tokens or
    concept names), normalized means lemmatized in this case"""
    stopwords: Optional[set] = None
    """If None the default set of stowords from spacy will be used.
    This must be a Set.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    max_document_length: int = 1000000
    """Documents longer  than this will be trimmed.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """


class CDBMaker(SerialisableBaseModel):
    """The Context Database (CDB) making part of the config"""
    name_versions: list = ['LOWER', 'CLEAN']
    """Name versions to be generated."""
    multi_separator: str = '|'
    """If multiple names or type_ids for a concept present in one row of a CSV,
    they are separated by the specified character."""
    remove_parenthesis: int = 5
    """Should preferred names with parenthesis be cleaned 0 means no,
    else it means if longer than or equal
    e.g. Head (Body part) -> Head"""
    min_letters_required: int = 2
    """Minimum number of letters required in a name to be accepted
    for a concept"""


class Ner(ComponentConfig):
    """The NER part of the config"""
    min_name_len: int = 3
    """Do not detect names below this limit, skip them"""
    max_skip_tokens: int = 2
    """When checking tokens for concepts you can have skipped tokens between
    used ones (usually spaces, new lines etc). This number tells you how many
    skipped can you have."""
    check_upper_case_names: bool = False
    """Check uppercase to distinguish uppercase and lowercase words that have
    a different meaning."""
    upper_case_limit_len: int = 4
    """Any name shorter than this must be uppercase in the text to be
    considered. If it is not uppercase
    it will be skipped."""
    try_reverse_word_order: bool = False
    """Try reverse word order for short concepts (2 words max),
    e.g. heart disease -> disease heart"""
    custom_cnf: Optional[Any] = None
    """The custom config for the component."""

    class Config:
        extra = 'allow'


class AnnotationOutput(SerialisableBaseModel):
    """The annotation output part of the config"""
    context_left: int = -1
    context_right: int = -1
    lowercase_context: bool = True
    include_text_in_output: bool = False


# NOTE: this class should have an attribute for each
#       medcat.components.types.CoreComponentType
class Components(SerialisableBaseModel):
    tagging: ComponentConfig = ComponentConfig()
    token_normalizing: ComponentConfig = ComponentConfig()
    ner: Ner = Ner()
    linking: Linking = Linking()
    comp_order: list[str] = ['tagging', 'token_normalizing',
                             'ner', 'linking']
    addons: list[ComponentConfig] = []


class TrainingDescriptor(SerialisableBaseModel):
    train_time_start: datetime
    train_time_end: datetime
    project_name: Optional[str]
    num_docs: int
    num_epochs: int = 1


T = TypeVar('T')
C = TypeVar('C', bound=Iterable)


class ModelMeta(SerialisableBaseModel):
    description: str = 'N/A'
    ontology: list[str] = []
    location: str = 'N/A'
    hash: str = ''  # TODO - implement
    history: list[str] = Field(default_factory=list)
    last_saved: datetime = Field(default_factory=datetime.now)
    unsup_trained: list[TrainingDescriptor] = []  # TODO - implement
    sup_trained: list[TrainingDescriptor] = []  # TODO - implement
    medcat_version: str = ''
    saved_environ: Environment = Field(default_factory=get_environment_info)

    def mark_saved_now(self):
        self.last_saved = datetime.now()
        self.saved_environ = get_environment_info()
        self.medcat_version = medcat_version

    # NOTE: this is expected to be called when training finished
    def add_unsup_training(self, start_time: datetime, num_docs: int,
                           num_epochs: int = 1, project_name: str = 'N/A'):
        """Add unsupervised training information based on data.

        NOTE: This will mark down the time taken for training by comparing
              the start time to the current time.

        Args:
            start_time (datetime): The time at which the training was started.
            num_docs (int): The number of documents trained.
            num_epochs (int, optional): The number of epochs. Defaults to 1.
            project_name (str, optional): The project name. Defaults to 'N/A'.
        """
        self.unsup_trained.append(TrainingDescriptor(
            train_time_start=start_time, train_time_end=datetime.now(),
            project_name=project_name, num_docs=num_docs,
            num_epochs=num_epochs))

    def add_sup_training(self, start_time: datetime, num_docs: int,
                         project_name: str) -> None:
        """Add supervised training information based on data.

        NOTE: This will mark down the time taken for training by comparing
              the start time to the current time.

        NOTE: This will be called for every project being trained separately.
              So if there's a MCT export being trained with multiple projects,
              multiple different training instances will be recorded.

        Args:
            start_time (datetime): The time at which the training was started.
            num_docs (int): The number of documents that were trained.
            project_name (str): The project name.
        """
        self.sup_trained.append(TrainingDescriptor(
            train_time_start=start_time, train_time_end=datetime.now(),
            project_name=project_name, num_docs=num_docs, num_epochs=1
        ))

    @contextmanager
    def prepare_and_report_training(self,
                                    data_iterator: C,
                                    num_epochs: int,
                                    supervised: bool = False,
                                    project_name: str = 'N/A'
                                    ) -> Iterator[C]:
        """Context manager for preparing training.

        This is used so that we can get the number of items in the data
        during training.

        Args:
            data_iterator (C): The data to be trained.
            num_epochs (int): The number of epochs to be used.
            supervised (bool, optional): Whether training is supervised.
                Defaults to False.
            project_name (str, optional): The project name. Defaults to 'N/A'.

        Yields:
            Iterator[C]: The same data that was input.
        """
        _names, _counts = [], [0]  # NOTE: 0 count for fallback

        def callback(name: str, count: int) -> None:
            _names.append(name)
            _counts.append(count)
        wrapped = callback_iterator(f"TRAIN-{id(data_iterator)}",
                                    data_iterator, callback)
        start_time = datetime.now()
        try:
            yield cast(C, wrapped)
        finally:
            # even if something fails, log the count
            num_docs = _counts[1]
            if supervised:
                self.add_sup_training(start_time=start_time,
                                      num_docs=num_docs,
                                      project_name=project_name)
            else:
                self.add_unsup_training(start_time=start_time,
                                        num_docs=num_docs,
                                        num_epochs=num_epochs,
                                        project_name=project_name)
            if len(_names) != 1:
                logger.warning(
                    "Something went wrong during %ssupervised training. "
                    "The number of documents trained was unable to be "
                    "clearly obtained. Counted %d names (%s) at %s",
                    'un' if not supervised else '', len(_names), _names,
                    _counts)


class Config(SerialisableBaseModel):
    general: General = General()
    components: Components = Components()
    # linking: Linking = Linking()
    preprocessing: Preprocessing = Preprocessing()
    cdb_maker: CDBMaker = CDBMaker()
    # ner: Ner = Ner()
    annotation_output: AnnotationOutput = AnnotationOutput()
    meta: ModelMeta = Field(default_factory=ModelMeta)


def get_important_config_parameters(config: Config) -> dict[str, Any]:
    return {
        "config.ponents.ner.min_name_len": {
            'value': config.components.ner.min_name_len,
            'description': ("Minimum detection length (found terms/mentions "
                            "shorter than this will not be detected).")
            },
        "config.ponents.ner.upper_case_limit_len": {
            'value': config.components.ner.upper_case_limit_len,
            'description': ("All detected terms shorter than this value have "
                            "to be uppercase, otherwise they will be ignored.")
            },
        "config.ponents.linking.similarity_threshold": {
            'value': config.components.linking.similarity_threshold,
            'description': ("If the confidence of the model is lower than "
                            "this a detection will be ignore.")
            },
        "config.ponents.linking.filters.cuis": {
            'value': len(config.components.linking.filters.cuis),
            'description': ("Length of the CUIs filter to be included in "
                            "outputs. If this is not 0 (i.e. not empty) its "
                            "best to check what is included before using the "
                            "model")
        },
        "config.general.spell_check": {
            'value': config.general.spell_check,
            'description': "Is spell checking enabled."
            },
        "config.general.spell_check_len_limit": {
            'value': config.general.spell_check_len_limit,
            'description': "Words shorter than this will not be spell checked."
            },
    }



from typing import Optional, cast
import numpy as np

from medcat.cdb import CDB
from medcat.data.mctexport import (MedCATTrainerExport,
                                   MedCATTrainerExportProject,
                                   MedCATTrainerExportDocument)
from medcat.tokenizing.tokens import MutableDocument, MutableEntity


class TestTrainSplitter:
    MAX_TEST_FRACTION = 0.3
    MIN_CNT_FOR_TEST = 10

    def __init__(self, data: MedCATTrainerExport, cdb: CDB,
                 test_size: float = 0.2):
        self.data = data
        self.cdb = cdb
        self.test_size = test_size
        self._reset()

    def _reset(self) -> None:
        self.cnts: dict[str, int] = {}
        self.test_cnts: dict[str, int] = {}
        self.total_anns = 0
        self.test_anns = 0
        self.test_prob = 0.90

    def _count_project(self, project: MedCATTrainerExportProject) -> None:
        cui_filter = None

        if 'cuis' in project and len(project['cuis'].strip()) > 0:
            cui_filter = [x.strip() for x in project['cuis'].split(",")]

        for ann in (ann for doc in project['documents']
                    for ann in doc['annotations']):
            if cui_filter is not None and ann['cui'] not in cui_filter:
                continue
            if ann['cui'] in self.cnts:
                self.cnts[ann['cui']] += 1
            else:
                self.cnts[ann['cui']] = 1

            self.total_anns += 1

    def split(self) -> tuple[MedCATTrainerExport, MedCATTrainerExport,
                             int, int]:
        # Count all CUIs
        for project in self.data['projects']:
            self._count_project(project)

        test_set: MedCATTrainerExport = {'projects': []}
        train_set: MedCATTrainerExport = {'projects': []}

        perm_arr: list[int] = cast(
            list[int],
            np.random.permutation(range(
                len(self.data['projects']))).tolist())

        for i_project in perm_arr:
            project = self.data['projects'][i_project]
            cui_filter = None

            # copy everything else, but reset documents list
            test_project: MedCATTrainerExportProject = project.copy()
            train_project: MedCATTrainerExportProject = project.copy()
            test_project['documents'] = []
            train_project['documents'] = []

            if 'cuis' in project and len(project['cuis'].strip()) > 0:
                cui_filter = [x.strip() for x in project['cuis'].split(",")]

            num_of_docs = len(project['documents'])
            for i_document in np.random.permutation(range(0, num_of_docs)):
                # Do we have enough documents in the test set
                if self.test_anns / self.total_anns >= self.test_size:
                    continue
                document = project['documents'][i_document]
                self._split_doc_train_test(document, cui_filter,
                                           train_project, test_project)

            test_set['projects'].append(test_project)
            train_set['projects'].append(train_project)

        return train_set, test_set, self.test_anns, self.total_anns

    def _split_doc_train_test(self, document: MedCATTrainerExportDocument,
                              cui_filter: Optional[list[str]],
                              train_project: MedCATTrainerExportProject,
                              test_project: MedCATTrainerExportProject):
        # Count CUIs for this document
        _cnts: dict[str, int] = {}
        doc_annotations = document['annotations']

        for ann in doc_annotations:
            if (cui_filter is not None and ann['cui'] not in cui_filter):
                continue
            if ann['cui'] in _cnts:
                _cnts[ann['cui']] += 1
            else:
                _cnts[ann['cui']] = 1

        is_test = self._should_add_to_test(_cnts)

        # Add to test set
        if is_test and np.random.rand() < self.test_prob:
            test_project['documents'].append(document)
            doc_annotations = document['annotations']

            for ann in doc_annotations:
                if cui_filter is not None and ann['cui'] not in cui_filter:
                    continue
                self.test_anns += 1
                if ann['cui'] in self.test_cnts:
                    self.test_cnts[ann['cui']] += 1
                else:
                    self.test_cnts[ann['cui']] = 1
        else:
            train_project['documents'].append(document)

    def _should_add_to_test(self, _cnts: dict[str, int]) -> bool:
        # Did we get more than 30% of concepts for any CUI with >=10 cnt
        return any(
            self.cnts[cui] >= self.MIN_CNT_FOR_TEST and
            (v + self.test_cnts.get(cui, 0)
             ) / self.cnts[cui] < self.MAX_TEST_FRACTION
            for cui, v in _cnts.items()
        )


def make_mc_train_test(data: MedCATTrainerExport,
                       cdb: CDB, test_size: float = 0.2) -> tuple:
    """Make train set.

    This is a disaster.

    Args:
        data (MedCATTrainerExport): The data.
        cdb (CDB): The concept database.
        test_size (float): The test size. Defaults to 0.2.

    Returns:
        tuple:
            The train set, the test set, the test annotations,
            and the total annotations
    """
    return TestTrainSplitter(data, cdb, test_size).split()


def get_false_positives(doc: MedCATTrainerExportDocument,
                        spacy_doc: MutableDocument
                        ) -> list[MutableEntity]:
    """Get the false positives within a trainer export.

    Args:
        doc (MedCATTrainerExportDocument): The trainer export.
        spacy_doc (MutableDocument): The annotated document.

    Returns:
        list[MutableEntity]: The list of false positive entities.
    """
    truth = set([(ent['start'], ent['cui']) for ent in doc['annotations']])

    fps = []
    for ent in spacy_doc.ner_ents:
        if (ent.base.start_index, ent.cui) not in truth:
            fps.append(ent)

    return fps



from typing import Callable, Protocol, Any, runtime_checkable, Optional

from medcat.components.types import BaseComponent, MutableEntity
from medcat.utils.registry import Registry
from medcat.config.config import ComponentConfig
from medcat.cdb import CDB
from medcat.vocab import Vocab
from medcat.tokenizing.tokenizers import BaseTokenizer


@runtime_checkable
class AddonComponent(BaseComponent, Protocol):
    """Base/abstract addon component class."""
    NAME_PREFIX: str = "addon_"
    NAME_SPLITTER: str = "."
    config: ComponentConfig

    @property
    def addon_type(self) -> str:
        pass

    def is_core(self) -> bool:
        return False

    @classmethod
    def get_folder_name_for_addon_and_name(
            cls, addon_type: str, name: str) -> str:
        return (cls.NAME_PREFIX + addon_type +
                cls.NAME_SPLITTER + name)

    def get_folder_name(self) -> str:
        return self.get_folder_name_for_addon_and_name(
            self.addon_type, self.name)

    @property
    def full_name(self) -> str:
        return self.addon_type + self.NAME_SPLITTER + str(self.name)

    @property
    def include_in_output(self) -> bool:
        return False  # default to False

    def get_output_key_val(self, ent: MutableEntity
                           ) -> tuple[str, dict[str, Any]]:
        pass


AddonClass = Callable[[ComponentConfig, BaseTokenizer,
                      CDB, Vocab, Optional[str]], AddonComponent]


_DEFAULT_ADDONS: dict[str, tuple[str, str]] = {
    'meta_cat': ('medcat.components.addons.meta_cat.meta_cat',
                 'MetaCATAddon.create_new_component'),
    'rel_cat': ('medcat.components.addons.relation_extraction.rel_cat',
                'RelCATAddon.create_new_component')
}

# NOTE: type error due to non-concrete type
_ADDON_REGISTRY = Registry(AddonComponent, _DEFAULT_ADDONS)  # type: ignore


def register_addon(addon_name: str,
                   addon_cls: AddonClass) -> None:
    """Register a new addon.

    Args:
        addon_name (str): The addon name.
        addon_cls (AddonClass): The addon creator.
    """
    _ADDON_REGISTRY.register(addon_name, addon_cls)


def get_addon_creator(addon_name: str) -> AddonClass:
    """Get the creator for an addon.

    Args:
        addon_name (str): The name of the addonl

    Returns:
        AddonClass: The creator of the addon.
    """
    return _ADDON_REGISTRY.get_component(addon_name)


def create_addon(
        addon_name: str, cnf: ComponentConfig,
        tokenizer: BaseTokenizer, cdb: CDB, vocab: Vocab,
        model_load_path: Optional[str]) -> AddonComponent:
    """Create an addon of the specified name with the specified arguments.

    All the `*args`, and `**kwrags` are passed to the creator.

    Args:
        addon_name (str): The name of the addon.
        cnf (ComponentConfig): The addon config.
        tokenizer (BaseTokenizer): The base tokenizer to be passed to creator.
        cdb (CDB): The CDB to be passed to creator.
        vocab (Vocab): The Vocab to be passed to creator.
        model_load_path (Optional[str]): The optional model load path to be
            passed to creator.


    Returns:
        AddonComponent: The resulting / created addon.
    """
    return get_addon_creator(addon_name)(
        cnf, tokenizer, cdb, vocab, model_load_path)



from typing import Optional

import logging
from medcat.tokenizing.tokens import MutableDocument
from medcat.components.types import CoreComponentType, AbstractCoreComponent
from medcat.components.ner.vocab_based_annotator import maybe_annotate_name
from medcat.utils.import_utils import ensure_optional_extras_installed
from medcat.tokenizing.tokenizers import BaseTokenizer
from medcat.vocab import Vocab
from medcat.cdb import CDB
from medcat.config.config import ComponentConfig

from ahocorasick import Automaton
import medcat


# NOTE: the _ is converted to - automatically
_EXTRA_NAME = "dict-ner"


ensure_optional_extras_installed(medcat.__name__, _EXTRA_NAME)


logger = logging.getLogger(__name__)


class NER(AbstractCoreComponent):
    name = 'cat_dict_ner'

    def __init__(self, tokenizer: BaseTokenizer,
                 cdb: CDB) -> None:
        self.tokenizer = tokenizer
        self.cdb = cdb
        self.config = self.cdb.config
        self.automaton = Automaton()
        self._rebuild_automaton()

    def _rebuild_automaton(self):
        # NOTE: every time the CDB changes (is dirtied)
        #       this will be recalculated
        logger.info("Rebuilding NER automaton (Aho-Corasick)")
        self.automaton.clear()
        # NOTE: we do not need name info for NER - only for linking
        ignored_min_len = 0
        for name in self.cdb.name2info.keys():
            clean_name = name.replace(self.config.general.separator, " ")
            if clean_name in self.automaton:
                # no need to duplicate
                continue
            if len(clean_name) < self.config.components.ner.min_name_len:
                # ignore names that are too short
                ignored_min_len += 1
                continue
            self.automaton.add_word(clean_name, clean_name)
        logger.debug("Ignored %d due to being smaller than minimum "
                     "allowed length (%d)", ignored_min_len,
                     self.config.components.ner.min_name_len)
        self.automaton.make_automaton()

    def get_type(self) -> CoreComponentType:
        return CoreComponentType.ner

    def __call__(self, doc: MutableDocument) -> MutableDocument:
        """Detect candidates for concepts - linker will then be able
        to do the rest. It adds `entities` to the doc.entities and each
        entity can have the entity.link_candidates - that the linker
        will resolve.

        Args:
            doc (MutableDocument):
                Spacy document to be annotated with named entities.

        Returns:
            doc (MutableDocument):
                Spacy document with detected entities.
        """
        if self.cdb.has_changed_names:
            self.cdb._reset_subnames()
            self._rebuild_automaton()
        text = doc.base.text.lower()
        for end_idx, raw_name in self.automaton.iter(text):
            start_idx = end_idx - len(raw_name) + 1
            cur_tokens = doc.get_tokens(start_idx, end_idx)
            if not isinstance(cur_tokens, list):
                # NOTE: this shouldn't really happen since
                #       there should be no entities defined
                #       before the NER step.
                #       But we will (at least for now) still handler this
                cur_tokens = list(cur_tokens)
            if not cur_tokens:
                # NOTE: the most likely reason for this is when matching
                #       a substring (e.g an abreviation in a longer word).
                #       In that case, no spacy tokens will match. But we
                #       don't really want to catch `mi` (for myocardial
                #       infarction) in "family".
                continue
            preprocessed_name = raw_name.replace(
                ' ', self.config.general.separator)
            maybe_annotate_name(self.tokenizer, preprocessed_name, cur_tokens,
                                doc, self.cdb, self.config)
        return doc

    @classmethod
    def create_new_component(
            cls, cnf: ComponentConfig, tokenizer: BaseTokenizer,
            cdb: CDB, vocab: Vocab, model_load_path: Optional[str]) -> 'NER':
        return cls(tokenizer, cdb)



from typing import Any
from contextlib import contextmanager
from pydantic import BaseModel


@contextmanager
def temp_changed_config(config: BaseModel, target: str, value: Any):
    """Context manager to change the config temporarily (within).

    Args:
        config (BaseModel): The config in question.
        target (str): The attribute name to change.
        value (Any): The temporary value to use.

    Raises:
        IllegalConfigPathException: If no previous value is available.
    """
    try:
        prev_value = getattr(config, target)
    except AttributeError as e:
        raise IllegalConfigPathException(target) from e
    setattr(config, target, value)
    try:
        yield
    finally:
        setattr(config, target, prev_value)


class IllegalConfigPathException(ValueError):

    def __init__(self, target_path: str):
        super().__init__(
            f"Config has no target path: {target_path}")



import torch
import logging
import os
import pickle
from typing import Any
import random

from medcat.components.addons.relation_extraction.tokenizer import (
    BaseTokenizerWrapper)
from medcat.config.config_rel_cat import ConfigRelCAT

from torch import nn


logger = logging.getLogger(__name__)


def split_list_train_test_by_class(data: list, sample_limit: int = -1,
                                   test_size: float = 0.2,
                                   shuffle: bool = True) -> tuple[list, list]:
    """

    Args:
        data (list): "output_relations": relation_instances,
            see create_base_relations_from_doc/csv for data columns
        sample_limit (int): Limit the number of samples per class,
            useful for dataset balancing . Defaults to -1.
        test_size (float): Defaults to 0.2.
        shuffle (bool): Shuffle data randomly. Defaults to True.

    Returns:
        tuple[list, list]: Train and test datasets
    """

    train_data = []
    test_data = []

    row_id_labels = {row_idx: data[row_idx][5] for row_idx in range(len(data))}
    lbl_id_to_name = {data[row_idx][5]: data[row_idx][4]
                      for row_idx in range((len(data)))}

    count_per_label = {lbl: list(row_id_labels.values()).count(
        lbl) for lbl in set(row_id_labels.values())}

    new_label_count_train = {}
    new_label_count_test = {}

    for lbl_id, count in count_per_label.items():
        if sample_limit != -1 and count > sample_limit:
            count = sample_limit

        _test_records_size = int(count * test_size)

        test_sample_count = 0
        train_sample_count = 0

        if _test_records_size not in [0, 1]:
            for row_idx, _lbl_id in row_id_labels.items():
                if _lbl_id == lbl_id:
                    if test_sample_count < _test_records_size:
                        test_data.append(data[row_idx])
                        test_sample_count += 1
                    else:
                        if sample_limit != -1:
                            if train_sample_count < sample_limit:
                                train_data.append(data[row_idx])
                                train_sample_count += 1
                        else:
                            train_data.append(data[row_idx])
                            train_sample_count += 1

        else:
            for row_idx, _lbl_id in row_id_labels.items():
                if _lbl_id == lbl_id:
                    train_data.append(data[row_idx])
                    test_data.append(data[row_idx])
                    train_sample_count += 1
                    test_sample_count += 1

        new_label_count_test[lbl_id] = test_sample_count
        new_label_count_train[lbl_id] = train_sample_count

    logging.info("Relations after train, test split :  train - %s | test - %s",
                 str(sum(new_label_count_train.values())),
                 str(sum(new_label_count_test.values())))

    for label_id in list(lbl_id_to_name.keys()):
        logging.info(" label: %s samples | train %s | test %s",
                     lbl_id_to_name[label_id],
                     str(new_label_count_train[label_id]),
                     str(new_label_count_test[label_id]))

    if shuffle:
        random.shuffle(train_data)
        random.shuffle(test_data)

    return train_data, test_data


def load_bin_file(file_name, path="./") -> Any:
    with open(os.path.join(path, file_name), 'rb') as f:
        data = pickle.load(f)
    return data


def save_bin_file(file_name, data, path="./"):
    with open(os.path.join(path, file_name), "wb") as f:
        pickle.dump(data, f)


def save_state(model, optimizer: torch.optim.AdamW,
               scheduler: torch.optim.lr_scheduler.MultiStepLR,
               epoch: int = 1, best_f1: float = 0.0, path: str = "./",
               model_name: str = "BERT", task: str = "train",
               is_checkpoint=False, final_export=False) -> None:
    """ Used by RelCAT.save() and RelCAT.train()
        Saves the RelCAT model state.
        For checkpointing multiple files are created, best_f1, loss etc. score.
        If you want to export the model after training set final_export=True
        and leave is_checkpoint=False.

    Args:
        model (BaseModel): BertMode | LlamaModel etc.
        optimizer (torch.optim.AdamW, optional): Defaults to None.
        scheduler (torch.optim.lr_scheduler.MultiStepLR, optional):
            Defaults to None.
        epoch (int): Defaults to None.
        best_f1 (float): Defaults to None.
        path (str):Defaults to "./".
        model_name (str): . Defaults to "BERT". This is used to
            checkpointing only.
        task (str): Defaults to "train". This is used to checkpointing only.
        is_checkpoint (bool): Defaults to False.
        final_export (bool): Defaults to False, if True then is_checkpoint
            must be False also. Exports model.state_dict(),
            out into "model.dat".
    """

    model_name = model_name.replace("/", "_")
    file_name = "%s_checkpoint_%s.dat" % (task, model_name)

    if not is_checkpoint:
        file_name = "%s_best_%s.dat" % (task, model_name)
        if final_export:
            file_name = "model.dat"
            torch.save(model.state_dict(), os.path.join(path, file_name))

    if is_checkpoint:
        torch.save({
            'epoch': epoch,
            'state_dict': model.state_dict(),
            'best_f1':  best_f1,
            'optimizer': optimizer.state_dict(),
            'scheduler': scheduler.state_dict()
        }, os.path.join(path, file_name))


def load_state(model, optimizer, scheduler, path: str = "./",
               model_name: str = "BERT", file_prefix: str = "train",
               load_best: bool = False,
               relcat_config: ConfigRelCAT = ConfigRelCAT()
               ) -> tuple[int, int]:
    """ Used by RelCAT.load() and RelCAT.train()

    Args:
        model (RelExtrBaseModel): RelExtrBaseModel,
            it has to be initialized before calling this method via
            RelExtr(Bert/Llama)Model(...)
        optimizer (_type_): optimizer
        scheduler (_type_): scheduler
        path (str, optional): Defaults to "./".
        model_name (str, optional): Defaults to "BERT".
        file_prefix (str, optional): Defaults to "train".
        load_best (bool, optional): Defaults to False.
        relcat_config (ConfigRelCAT): Defaults to ConfigRelCAT().

    Returns:
        tuple (int, int): last epoch and f1 score.
    """

    device: torch.device = torch.device(relcat_config.general.device)

    model_name = model_name.replace("/", "_")
    logging.info("Attempting to load RelCAT model on device: " + str(device))
    checkpoint_path = os.path.join(
        path, file_prefix + "_checkpoint_%s.dat" % model_name)
    best_path = os.path.join(
        path, file_prefix + "_best_%s.dat" % model_name)
    start_epoch, best_f1, checkpoint = 0, 0, None

    if load_best is True and os.path.isfile(best_path):
        checkpoint = torch.load(best_path, map_location=device)
        logging.info("Loaded best model.")
    elif os.path.isfile(checkpoint_path):
        checkpoint = torch.load(checkpoint_path, map_location=device)
        logging.info("Loaded checkpoint model.")

    if checkpoint is not None:
        start_epoch = checkpoint['epoch']
        best_f1 = checkpoint['best_f1']
        model.load_state_dict(checkpoint['state_dict'])
        model.to(device)

        if optimizer is None:
            parameters = filter(lambda p: p.requires_grad, model.parameters())
            optimizer = torch.optim.AdamW(
                params=parameters, lr=relcat_config.train.lr,
                weight_decay=relcat_config.train.adam_weight_decay,
                betas=relcat_config.train.adam_betas,
                eps=relcat_config.train.adam_epsilon)

        if scheduler is None:
            scheduler = torch.optim.lr_scheduler.MultiStepLR(
                optimizer, milestones=relcat_config.train.multistep_milestones,
                gamma=relcat_config.train.multistep_lr_gamma)
        optimizer.load_state_dict(checkpoint['optimizer'])
        scheduler.load_state_dict(checkpoint['scheduler'])
        logging.info("Loaded model and optimizer.")

    return start_epoch, best_f1


def save_results(data, model_name: str = "BERT", path: str = "./",
                 file_prefix: str = "train"):
    save_bin_file(file_prefix + "_losses_accuracy_f1_per_epoch_%s.dat" %
                  model_name, data, path)


def load_results(path: str, model_name: str = "BERT",
                 file_prefix: str = "train"
                 ) -> tuple[list, list, list]:
    data_dict_path = os.path.join(
        path, file_prefix + f"_losses_accuracy_f1_per_epoch_{model_name}.dat")

    data_dict: dict = {"losses_per_epoch": [],
                       "accuracy_per_epoch": [], "f1_per_epoch": []}
    if os.path.isfile(data_dict_path):
        data_dict = load_bin_file(data_dict_path)

    return (data_dict["losses_per_epoch"],
            data_dict["accuracy_per_epoch"],
            data_dict["f1_per_epoch"])


def create_tokenizer_pretrain(tokenizer: BaseTokenizerWrapper,
                              relcat_config: ConfigRelCAT
                              ) -> BaseTokenizerWrapper:
    """This method simply adds the default special tokens that we ecounter.

    Args:
        tokenizer (BaseTokenizerWrapper):
            BERT/Llama tokenizer.
        relcat_config (ConfigRelCAT): The RelCAT config.

    Returns:
        BaseTokenizerWrapper: The same tokenizer.
    """
    special_tags = (
        relcat_config.general.tokenizer_relation_annotation_special_tokens_tags
    )
    tokenizer.hf_tokenizers.add_tokens(
        special_tags, special_tokens=True)

    # used in llama tokenizer, may produce issues with other tokenizers
    tokenizer.hf_tokenizers.add_special_tokens(
        relcat_config.general.tokenizer_other_special_tokens)

    return tokenizer


def create_dense_layers(relcat_config: ConfigRelCAT):

    # dense layers
    fc1 = nn.Linear(
        relcat_config.model.model_size,
        relcat_config.model.hidden_size)
    fc2 = nn.Linear(
        relcat_config.model.hidden_size,
        int(relcat_config.model.hidden_size / 2))
    fc3 = nn.Linear(
        int(relcat_config.model.hidden_size / 2),
        relcat_config.train.nclasses)

    return fc1, fc2, fc3


def get_annotation_schema_tag(sequence_output: torch.Tensor,
                              input_ids: torch.Tensor, special_tag: list
                              ) -> torch.Tensor:
    """ Gets to token sequences from the sequence_ouput for the specific token
        tag ids in self.relcat_config.general.annotation_schema_tag_ids.

    Args:
        sequence_output (torch.Tensor): hidden states/embeddings for
            each token in the input text
        input_ids (torch.Tensor): input token ids
        special_tag (list): special annotation token id pairs

    Returns:
        torch.Tensor: new seq_tags
    """

    # returns: row ids, idx of token[0]/star token in row
    idx_start = torch.where(input_ids == special_tag[0])
    # returns: row ids, idx of token[1]/end token in row
    idx_end = torch.where(input_ids == special_tag[1])

    seen = []  # List to store seen elements and their indices
    duplicate_indices = []

    for i in range(len(idx_start[0])):
        if idx_start[0][i] in seen:
            duplicate_indices.append(i)
        else:
            seen.append(idx_start[0][i])

    if len(duplicate_indices) > 0:
        logger.info("Duplicate entities found, removing them...")
        for idx_remove in duplicate_indices:
            idx_start_0 = torch.cat((idx_start[0][:idx_remove],
                                     idx_start[0][idx_remove + 1:]))
            idx_start_1 = torch.cat((idx_start[1][:idx_remove],
                                     idx_start[1][idx_remove + 1:]))
            idx_start = (idx_start_0, idx_start_1)

    seen = []
    duplicate_indices = []

    for i in range(len(idx_end[0])):
        if idx_end[0][i] in seen:
            duplicate_indices.append(i)
        else:
            seen.append(idx_end[0][i])

    if len(duplicate_indices) > 0:
        logger.info("Duplicate entities found, removing them...")
        for idx_remove in duplicate_indices:
            idx_end_0 = torch.cat((idx_end[0][:idx_remove],
                                   idx_end[0][idx_remove + 1:]))
            idx_end_1 = torch.cat((idx_end[1][:idx_remove],
                                   idx_end[1][idx_remove + 1:]))
            idx_end = (idx_end_0, idx_end_1)

    assert len(idx_start[0]) == input_ids.shape[0]
    assert len(idx_start[0]) == len(idx_end[0])

    sequence_output_entities: list[torch.Tensor] = []

    for i in range(len(idx_start[0])):
        to_append = sequence_output[i, idx_start[1][i] + 1:idx_end[1][i], ]

        to_append, _ = torch.max(to_append, axis=0)  # type: ignore

        sequence_output_entities.append(to_append)
    return torch.stack(sequence_output_entities)



import logging
import sys


logger = logging.getLogger(__name__)


def has_spacy_model(model_name: str) -> bool:
    """Checks if the spacy model is available.

    Args:
        model_name (str): The model name.

    Returns:
        bool: True if the model is available, False otherwise.
    """
    import spacy.util
    return model_name in spacy.util.get_installed_models()


def ensure_spacy_model(model_name: str) -> None:
    """Ensure the specified spacy model exists.

    If the model does not currently exist, it will attempt downloading it.

    Args:
        model_name (str): The spacy model name.
    """
    if has_spacy_model(model_name):
        return
    import subprocess
    # running in subprocess so that we can catch the exception
    # if the model name is unknown. Otherwise we'd just be bumped
    # out of python (sys.exit).
    cmd = f"{sys.executable} -m spacy download {model_name}"
    logger.info("Installing the spacy model %s using the CLI command "
                "'%s'", model_name, cmd)
    subprocess.run(cmd.split(" "), check=True)



import os
from typing import Optional
from multiprocessing import cpu_count
from functools import lru_cache
import logging


DEFAULT_SPACY_MODEL = 'en_core_web_md'
DEFAULT_PACK_NAME = "medcat2_model_pack"
COMPONENTS_FOLDER = "saved_components"
AVOID_LEGACY_CONVERSION_ENVIRON = "MEDCAT_AVOID_LECACY_CONVERSION"


def avoid_legacy_conversion() -> bool:
    return os.environ.get(
        AVOID_LEGACY_CONVERSION_ENVIRON, "False").lower() == "true"


class LegacyConversionDisabledError(Exception):
    """Raised when legacy conversion is disabled."""

    def __init__(self, component_name: str):
        super().__init__(
            f"Legacy conversion is disabled (while loading {component_name}). "
            f"Set the environment variable {AVOID_LEGACY_CONVERSION_ENVIRON} "
            "to `False` to allow conversion.")


def doing_legacy_conversion_message(
        logger: logging.Logger, component_name: str, file_path: str = '',
        level: int = logging.WARNING
        ) -> None:
    logger.log(
        level,
        "Doing legacy conversion on %s (at '%s'). "
        "Set the environment variable %s "
        "to `True` to avoid this.",
        component_name, file_path, AVOID_LEGACY_CONVERSION_ENVIRON)


@lru_cache(maxsize=100)
def default_weighted_average(step: int, factor: float = 0.0004) -> float:
    return max(0.1, 1 - (step ** 2 * factor))


def workers(workers_override: Optional[int] = None) -> int:
    """Get number of workers.

    Either the number of workers specified (if done so).
    Or the number of workers available (i.e cpu count - 1).

    Args:
        workers_override (Optional[int], optional):
            The number of workers to use. Defaults to None.

    Returns:
        int: _description_
    """
    if workers_override is not None:
        return workers_override
    return max(cpu_count() - 1, 1)


class StatusTypes:
    PRIMARY_STATUS_NO_DISAMB = 'P'
    PRIMARY_STATUS_W_DISAMB = 'PD'
    PRIMARY_STATUS: set[str] = {PRIMARY_STATUS_NO_DISAMB,
                                PRIMARY_STATUS_W_DISAMB}
    MUST_DISAMBIGATE = 'N'
    AUTOMATIC = 'A'
    ALLOWED_STATUS = {PRIMARY_STATUS_NO_DISAMB, MUST_DISAMBIGATE, AUTOMATIC}
    DO_DISAMBUGATION = {MUST_DISAMBIGATE, PRIMARY_STATUS_W_DISAMB}



from typing import Any, Iterator, Optional
import yaml
import json
import logging
import tqdm
import datetime
import os

from pydantic import BaseModel, Field

from medcat.cat import CAT
from medcat.utils.regression.targeting import TranslationLayer, OptionSet
from medcat.utils.regression.targeting import (
    FinalTarget, TargetedPhraseChanger)
from medcat.utils.regression.utils import (
    partial_substitute, MedCATTrainerExportConverter)
from medcat.utils.regression.utils import pick_random_edits, EditGetter
from medcat.utils.regression.results import (
    MultiDescriptor, ResultDescriptor, Finding)
from medcat.components.normalizing.normalizer import BasicSpellChecker

logger = logging.getLogger(__name__)


class RegressionCase(BaseModel):
    """A regression case that has a name, defines options, filters and phrases.
    """
    name: str
    options: OptionSet
    phrases: list[str]
    report: ResultDescriptor

    def check_specific_for_phrase(self, cat: CAT, target: FinalTarget,
                                  translation: TranslationLayer
                                  ) -> tuple[Finding, Optional[str]]:
        """Checks whether the specific target along with the specified phrase
        is able to be identified using the specified model.

        Args:
            cat (CAT): The model
            target (FinalTarget): The final target configuration
            translation (TranslationLayer): The translation layer

        Raises:
            MalformedRegressionCaseException:
                If there are too many placeholders in phrase.

        Returns:
            tuple[Finding, Optional[str]]:
                The nature to which the target was (or wasn't) identified
        """
        phrase, cui, name, placeholder = (
            target.final_phrase, target.cui, target.name, target.placeholder)
        nr_of_placeholders = phrase.count(placeholder)
        if nr_of_placeholders != 1:
            raise MalformedRegressionCaseException(
                f"Got {nr_of_placeholders} placeholders "
                f"({placeholder}) (expected 1) for phrase: " +
                phrase)
        ph_start = phrase.find(placeholder)
        res = cat.get_entities(
            phrase.replace(placeholder, name), only_cui=False)
        ents = res['entities']
        finding = Finding.determine(cui, ph_start, ph_start + len(name),
                                    translation, ents)
        if finding is Finding.IDENTICAL:
            logger.debug(
                'Matched test case %s in phrase "%s"', (cui, name), phrase)
        else:
            found_cuis = [ents[nr]['cui'] for nr in ents]
            found_names = [ents[nr]['source_value'] for nr in ents]
            cuis_names = ', '.join([f'{fcui}|{fname}'
                                    for fcui, fname
                                    in zip(found_cuis, found_names)])
            logger.debug(
                'FAILED to (fully) match (%s) test case %s in phrase "%s", '
                'found the following CUIS/names: %s',
                finding, (cui, name), phrase, cuis_names)
        self.report.report(target, finding)
        return finding

    def estimate_num_of_diff_subcases(self) -> int:
        return len(self.phrases) * self.options.estimate_num_of_subcases()

    def get_distinct_cases(self, translation: TranslationLayer,
                           edit_distance: tuple[int, int, int],
                           use_diacritics: bool
                           ) -> Iterator[Iterator[FinalTarget]]:
        """Gets the various distinct sub-case iterators.

        The sub-cases are those that can be determine without the translation
        layer. However, the translation layer is included here since it
        streamlines the operation.

        Args:
            translation (TranslationLayer): The translation layer.
            edit_distance (tuple[int, int, int]): The edit distance(s) to try.
            use_diacritics (bool): Whether to use diacritics for edit distance.

        Yields:
            Iterator[Iterator[FinalTarget]]:
                The iterator of iterators of different sub cases.
        """
        # for each phrase and for each placeholder based option
        for changer in self.options.get_preprocessors_and_targets(translation):
            for phrase in self.phrases:
                yield self._get_subcases(phrase, changer, translation,
                                         edit_distance, use_diacritics)

    def _get_subcases(self, phrase: str, changer: TargetedPhraseChanger,
                      translation: TranslationLayer,
                      edit_distance: tuple[int, int, int],
                      use_diacritics: bool,
                      ) -> Iterator[FinalTarget]:
        cui, placeholder = changer.cui, changer.placeholder
        changed_phrase = changer.changer(phrase)
        edit_dist, edit_rn_seed, edit_pick = edit_distance
        if edit_dist >= 2 or edit_dist < 0:
            raise ValueError(
                "Edit distance above 2 or below 0 are unsupported")
        if edit_dist:
            get_edits: EditGetter
            if edit_dist == 1:
                get_edits = BasicSpellChecker.raw_edits1
            elif edit_dist == 2:
                get_edits = BasicSpellChecker.raw_edits2
            else:
                raise ValueError(
                    "Edit distance above 2 or below 0 are unsupported")
        for raw_name in translation.get_names_of(cui, changer.onlyprefnames):
            name_variant = 0
            if edit_dist:
                name_gen = get_edits(
                    raw_name, use_diacritics, True)
                all_names = list(
                    pick_random_edits(name_gen, edit_pick, len(raw_name),
                                      edit_dist, edit_rn_seed))
            else:
                all_names = [raw_name]
            for name in all_names:
                if edit_dist:
                    logger.debug(
                        "Changed name from '%s' to '%s' (variant %d, edit "
                        "distance %s, seed %d, picking %d)", raw_name, name,
                        name_variant, edit_dist, edit_rn_seed, edit_pick)
                    name_variant += 1
                num_of_phs = changed_phrase.count(placeholder)
                if num_of_phs == 1:
                    yield FinalTarget(
                        placeholder=placeholder,
                        cui=cui, name=name,
                        final_phrase=changed_phrase)
                    continue
                for cntr in range(num_of_phs):
                    final_phrase = partial_substitute(
                        changed_phrase, placeholder, name, cntr)
                    yield FinalTarget(
                        placeholder=placeholder,
                        cui=cui, name=name,
                        final_phrase=final_phrase)

    def to_dict(self) -> dict:
        """Converts the RegressionCase to a dict for serialisation.

        Returns:
            dict: The dict representation
        """
        d: dict[str, Any] = {'phrases': list(self.phrases)}
        targeting = self.options.to_dict()
        d['targeting'] = targeting
        return d

    @classmethod
    def from_dict(cls, name: str, in_dict: dict) -> 'RegressionCase':
        """Construct the regression case from a dict.

        The expected structure:
        {
            'targeting': {
                [
                    # the placeholder to be replaced
                    'placeholder': '[DIAGNOSIS]'
                    'cuis': ['cui1', 'cui2']
                    'prefname-only': 'false', # optional
                ]
            },
            'phrases': ['phrase %s'] # possible multiple
        }

        Args:
            name (str): The name of the case
            in_dict (dict): The dict describing the case

        Raises:
            ValueError: If the input dict does not have the 'targeting' section
            ValueError: If there are no phrases defined

        Returns:
            RegressionCase: The constructed regression cases.
        """
        # set up targeting
        if 'targeting' not in in_dict:
            raise ValueError('Input dict should define targeting')
        targeting_section = in_dict['targeting']
        # set up options
        options = OptionSet.from_dict(targeting_section)
        # all_cases: list['RegressionCase'] = []
        # for option in options:
        #     # set up test phrases
        if 'phrases' not in in_dict:
            raise ValueError('Input dict should defined phrases')
        phrases = in_dict['phrases']
        if not isinstance(phrases, list):
            phrases = [phrases]  # just one defined
        if not phrases:
            raise ValueError('Need at least one target phrase')
        return RegressionCase(name=name, options=options,
                              phrases=phrases,
                              report=ResultDescriptor(name=name))

    def __hash__(self) -> int:
        return hash(str(self.to_dict()))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, RegressionCase):
            return False
        return self.to_dict() == other.to_dict()


UNKNOWN_METADATA = 'Unknown'


def get_ontology_and_version(model_card: dict) -> tuple[str, str]:
    """Attempt to get ontology (and its version) from a model card dict.

    If no ontology is found, 'Unknown' is returned.
    The version is always returned as the first source ontology.
    That is, unless the specified location does not exist in the model card,
    in which case 'Unknown' is returned.

    The ontology is assumed to be described at:
        model_card['Source Ontology'][0] (or model_card['Source Ontology']
        if it's a string instead of a list)

    The ontology version is read from:
        model_card['Source Ontology'][0] (or model_card['Source Ontology']
        if it's a string instead of a list)

    Currently, only SNOMED-CT, UMLS and ICD are supported / found.

    Args:
        model_card (dict): The input model card.

    Returns:
        tuple[str, str]:
            The ontology (if found) or 'Unknown';
            and the version (if found) or 'Unknown'
    """
    try:
        ont_list = model_card['Source Ontology']
        if isinstance(ont_list, list):
            ont1 = ont_list[0]
        elif isinstance(ont_list, str):
            ont1 = ont_list
        else:
            raise KeyError(f"Unknown source ontology: {ont_list}")
    except KeyError as key_err:
        logger.warning(
            "Didn't find the expected source ontology from the model card!",
            exc_info=key_err)
        return UNKNOWN_METADATA, UNKNOWN_METADATA
    # find ontology
    if 'SNOMED' in ont1.upper():
        return 'SNOMED-CT', ont1
    elif 'UMLS' in ont1.upper():
        return 'UMLS', ont1
    elif 'ICD' in ont1.upper():
        return 'ICD', ont1
    else:
        return UNKNOWN_METADATA, ont1


class MetaData(BaseModel):
    """The metadata for the regression suite.

    This should define which ontology (e.g UMLS or SNOMED) as well as
    which version was used when generating the regression suite.

    The metadata may contain further information as well, this may include
    the annotator(s) involved when converting from MCT export or other
    relevant data.
    """
    ontology: str
    ontology_version: str
    extra: dict = {}
    regr_suite_creation_date: str = Field(
        default_factory=lambda: str(datetime.datetime.now()))

    @classmethod
    def from_modelcard(cls, model_card: dict) -> 'MetaData':
        """Generate a MetaData object from a model card.

        This involves reading ontology info and version from the model card.

        It must be noted that the model card should be provided as a dict
        not a string.

        Args:
            model_card (dict): The CAT modelcard

        Returns:
            MetaData: The resulting MetaData
        """
        ontology, ont_version = get_ontology_and_version(model_card)
        return MetaData(
            ontology=ontology, ontology_version=ont_version, extra=model_card)

    @classmethod
    def unknown(self) -> 'MetaData':
        return MetaData(
            ontology=UNKNOWN_METADATA, ontology_version=UNKNOWN_METADATA,
            extra={}, regr_suite_creation_date=UNKNOWN_METADATA)


def fix_np_float64(d: dict) -> None:
    """Fix numpy.float64 in dictionary for yaml saving purposes.

    These types of objects are unable to be cleanly serialized using yaml.
    So we need to convert them to the corresponding floats.

    The changes will be made within the dictionary itself
    as well as dictionaries within, recursively.

    Args:
        d (dict): The input dict
    """
    import numpy as np
    for k, v in d.items():
        if isinstance(v, np.float64):
            d[k] = float(v)
        if isinstance(v, dict):
            fix_np_float64(v)


class RegressionSuite:
    """The regression checker.
    This is used to check a bunch of regression cases at once against a model.

    Args:
        cases (list[RegressionCase]): The list of regression cases
        metadata (MetaData): The metadata for the regression suite
        use_report (bool): Whether or not to use the report functionality.
            Defaults to False.
    """

    def __init__(self, cases: list[RegressionCase],
                 metadata: MetaData, name: str) -> None:
        self.cases: list[RegressionCase] = cases
        self.report = MultiDescriptor(name=name)
        self.metadata = metadata
        for case in self.cases:
            self.report.parts.append(case.report)

    def get_all_distinct_cases(self, translation: TranslationLayer,
                               edit_distance: tuple[int, int, int],
                               use_diacritics: bool
                               ) -> Iterator[tuple[RegressionCase,
                                                   Iterator[FinalTarget]]]:
        """Gets all the distinct cases for this regression suite.

        While distinct cases can be determined without the translation layer,
        including it here simplifies the process.

        Args:
            translation (TranslationLayer): The translation layer.
            edit_distance (tuple[int, int, int]): The edit distance(s) to try.
                Defaults to (0, 0, 0).
            use_diacritics (bool): Whether to use diacritics for edit distance.

        Yields:
            Iterator[tuple[RegressionCase, Iterator[FinalTarget]]]:
                The generator of the regression case along with
                its corresponding sub-cases.
        """
        for regr_case in self.cases:
            for subcase in regr_case.get_distinct_cases(
                    translation, edit_distance, use_diacritics):
                yield regr_case, subcase

    def estimate_total_distinct_cases(self) -> int:
        return sum(rc.estimate_num_of_diff_subcases() for rc in self.cases)

    def iter_subcases(self, translation: TranslationLayer,
                      show_progress: bool = True,
                      edit_distance: tuple[int, int, int] = (0, 0, 0),
                      use_diacritics: bool = False,
                      ) -> Iterator[tuple[RegressionCase, FinalTarget]]:
        """Iterate over all the sub-cases.

        Each sub-case present a unique target (phrase, concept, name) on
        the corresponding regression case.

        Args:
            translation (TranslationLayer): The translation layer.
            show_progress (bool): Whether to show progress. Defaults to True.
            edit_distance (tuple[int, int, int]): The edit distance(s) to try.
                Defaults to (0, 0, 0).
            use_diacritics (bool): Whether to use diacritics for edit distance.

        Yields:
            Iterator[tuple[RegressionCase, FinalTarget]]: The generator of the
                regression case along with each of the final target sub-cases.
        """
        total = self.estimate_total_distinct_cases()
        for (regr_case, subcase) in tqdm.tqdm(
                self.get_all_distinct_cases(translation, edit_distance,
                                            use_diacritics),
                total=total, disable=not show_progress):
            for target in subcase:
                yield regr_case, target

    def check_model(self, cat: CAT, translation: TranslationLayer,
                    edit_distance: tuple[int, int, int] = (0, 0, 0),
                    use_diacritics: bool = False,
                    ) -> MultiDescriptor:
        """Checks model and generates a report

        Args:
            cat (CAT): The model to check against
            translation (TranslationLayer): The translation layer
            edit_distance (tuple[int, int, int]):
                The edit distance of the names. Defaults to (0, 0, 0).
            use_diacritics (bool): Whether to use diacritics for edit distance.

        Returns:
            MultiDescriptor: A report description
        """
        for regr_case, target in self.iter_subcases(
                translation, True, edit_distance, use_diacritics):
            # NOTE: the finding is reported in the per-case report
            regr_case.check_specific_for_phrase(cat, target, translation)
        return self.report

    def __str__(self) -> str:
        return f'RegressionTester[cases={self.cases}]'

    def __repr__(self) -> str:
        return f'<{self}>'

    def to_dict(self) -> dict:
        """Converts the RegressionChecker to dict for serialisation.

        Returns:
            dict: The dict representation
        """
        d = {}
        for case in self.cases:
            d[case.name] = case.to_dict()
        d['meta'] = self.metadata.model_dump()
        fix_np_float64(d['meta'])

        return d

    def to_yaml(self) -> str:
        """Convert the RegressionChecker to YAML string.

        Returns:
            str: The YAML representation
        """
        return yaml.safe_dump(self.to_dict())

    def __eq__(self, other: object) -> bool:
        # only checks cases
        if not isinstance(other, RegressionSuite):
            return False
        return self.cases == other.cases

    @classmethod
    def from_dict(cls, in_dict: dict, name: str) -> 'RegressionSuite':
        """Construct a RegressionChecker from a dict.

        Most of the parsing is handled in RegressionChecker.from_dict.
        This just assumes that each key in the dict is a name
        and each value describes a RegressionCase.

        Args:
            in_dict (dict): The input dict.
            name (str): The name of the regression suite.

        Returns:
            RegressionChecker: The built regression checker
        """
        cases: list[RegressionCase] = []
        for case_name, details in in_dict.items():
            if case_name == 'meta':
                continue  # ignore metadata
            add_case = RegressionCase.from_dict(case_name, details)
            cases.append(add_case)
        if 'meta' not in in_dict:
            logger.warn("Loading regression suite without any meta data")
            metadata = MetaData.unknown()
        else:
            metadata = MetaData.parse_obj(in_dict['meta'])
        return RegressionSuite(cases=cases, metadata=metadata, name=name)

    @classmethod
    def from_yaml(cls, file_name: str) -> 'RegressionSuite':
        """Constructs a RegressionChcker from a YAML file.

        The from_dict method is used for the construction from the dict.

        Args:
            file_name (str): The file name

        Returns:
            RegressionChecker: The constructed regression checker
        """
        with open(file_name) as f:
            data = yaml.safe_load(f)
        return RegressionSuite.from_dict(
            data, name=os.path.basename(file_name))

    @classmethod
    def from_mct_export(cls, file_name: str) -> 'RegressionSuite':
        with open(file_name) as f:
            data = json.load(f)
        converted = MedCATTrainerExportConverter(data).convert()
        return RegressionSuite.from_dict(
            converted, name=os.path.basename(file_name))


class MalformedRegressionCaseException(ValueError):

    def __init__(self, *args: object) -> None:
        super().__init__(*args)



import logging
import os
import torch

from torch import nn
from transformers.models.bert.modeling_bert import BertModel

from medcat.config.config_rel_cat import ConfigRelCAT
from medcat.components.addons.relation_extraction.ml_utils import (
    create_dense_layers)
from medcat.components.addons.relation_extraction.models import (
    RelExtrBaseModel)
from medcat.components.addons.relation_extraction.bert.config import (
    RelExtrBertConfig)


logger = logging.getLogger(__name__)


class RelExtrBertModel(RelExtrBaseModel):
    """ BertModel class for RelCAT
    """

    name = "bertmodel_relcat"

    def __init__(self, pretrained_model_name_or_path: str,
                 relcat_config: ConfigRelCAT,
                 model_config: RelExtrBertConfig):
        """ Class to hold the BERT model + model_config

        Args:
            pretrained_model_name_or_path (str): path to load the model from,
                    this can be a HF model i.e: "bert-base-uncased",
                    if left empty, it is normally assumed that a model
                    is loaded from 'model.dat' using the RelCAT.load() method.
                    So if you are initializing/training a model from scratch
                    be sure to base it on some model.
            relcat_config (ConfigRelCAT): relcat config.
            model_config (Union[RelExtrBaseConfig | RelExtrBertConfig]):
                HF bert config for model.
        """
        super().__init__(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            relcat_config=relcat_config, model_config=model_config)

        self.relcat_config: ConfigRelCAT = relcat_config
        self.model_config: RelExtrBertConfig = model_config
        self.pretrained_model_name_or_path: str = pretrained_model_name_or_path

        self.hf_model = BertModel(model_config.hf_model_config)

        for param in self.hf_model.parameters():
            if self.relcat_config.model.freeze_layers:
                param.requires_grad = False
            else:
                param.requires_grad = True

        self.drop_out = nn.Dropout(self.relcat_config.model.dropout)

        # dense layers
        self.fc1, self.fc2, self.fc3 = create_dense_layers(self.relcat_config)

    @classmethod
    def load_specific(cls, pretrained_model_name_or_path: str,
                      relcat_config: ConfigRelCAT,
                      model_config: RelExtrBertConfig, **kwargs
                      ) -> "RelExtrBertModel":

        model = RelExtrBertModel(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            relcat_config=relcat_config, model_config=model_config)

        model_path = os.path.join(pretrained_model_name_or_path, "model.dat")

        if os.path.exists(model_path):
            model.load_state_dict(torch.load(
                model_path, map_location=relcat_config.general.device))
            logger.info("Loaded model from file: %s", model_path)
        elif pretrained_model_name_or_path:
            model.hf_model = BertModel.from_pretrained(
                pretrained_model_name_or_path=pretrained_model_name_or_path,
                config=model_config.hf_model_config,
                ignore_mismatched_sizes=True, **kwargs)
            logger.info("Loaded model from pretrained: %s",
                        pretrained_model_name_or_path)
        else:
            pretrained_model = cls.pretrained_model_name_or_path
            model.hf_model = BertModel.from_pretrained(
                pretrained_model_name_or_path=pretrained_model,
                config=model_config.hf_model_config,
                ignore_mismatched_sizes=True, **kwargs)
            logger.info("Loaded model from pretrained: %s", pretrained_model)

        return model



import dill
from typing import Optional, Dict
from transformers.tokenization_utils_base import PreTrainedTokenizerBase
import logging

logger = logging.getLogger(__name__)


class TransformersTokenizer:
    """Args:
        hf_tokenizer
            Must be able to return token offsets.
        max_len:
            Max sequence length, if longer it will be split into
            multiple examples.
        id2type:
            Can be ignored in most cases, should be a map from token to 'start'
            or 'sub' meaning is the token a subword or the start/full word.
            For BERT 'start' is everything that does not begin with ##.
        cui2name:
            Map from CUI to full name for labels.
    """

    def __init__(self,
                 hf_tokenizer: Optional[PreTrainedTokenizerBase] = None,
                 max_len: int = 512,
                 id2type: Optional[Dict] = None,
                 cui2name: Optional[Dict] = None) -> None:
        self.hf_tokenizer = hf_tokenizer
        self.max_len = max_len
        # We'll keep the 'X' in case id2type is provided
        self.label_map = {'O': 0, 'X': 1}
        self.id2type = id2type
        self.cui2name = cui2name

    def calculate_label_map(self, dataset) -> None:
        for cuis in dataset['ent_cuis']:
            for cui in cuis:
                if cui not in self.label_map:
                    self.label_map[cui] = len(self.label_map)

    def encode(self, examples: Dict, ignore_subwords: bool = False) -> Dict:
        """Used with huggingface datasets map function to convert medcat_ner
        dataset into the appropriate form for NER with BERT. It will split
        long text segments into max_len sequences (performs chunking).

        Args:
            examples (Dict):
                Stream of examples.
            ignore_subwords (bool):
                If set to `True` subwords of any token will get the special
                label `X`.

        Returns:
            Dict: The same dict, modified.
        """
        self.hf_tokenizer = self.ensure_tokenizer()
        old_ids = examples['id']
        old_names = examples['name']
        examples['input_ids'] = []
        examples['labels'] = []
        examples['id'] = []
        examples['name'] = []

        for _ind, example in enumerate(zip(
                examples['text'], examples['ent_starts'],
                examples['ent_ends'], examples['ent_cuis'])):
            tokens = self.hf_tokenizer(
                example[0], return_offsets_mapping=True,
                add_special_tokens=False)
            entities = [(start, end, cui)
                        for start, end, cui in
                        zip(example[1], example[2], example[3])]
            entities.sort(key=lambda x: x[0])
            input_ids = []
            labels = []

            tkn_part_of_entity = False
            for ind in range(len(tokens['offset_mapping'])):
                offset = tokens['offset_mapping'][ind]
                input_ids.append(tokens['input_ids'][ind])

                if (entities and (offset[0] >= entities[0][0] and
                                  offset[1] <= entities[0][1])):
                    # Means this token is part of entity at position 0
                    tkn_part_of_entity = True
                    if (not ignore_subwords or (
                            self.id2type is not None and
                            self.id2type[tokens['input_ids'][ind]] == 'start'
                            )):
                        labels.append(self.label_map[entities[0][2]])
                    else:
                        labels.append(self.label_map['X'])

                    if entities[0][1] <= offset[1]:
                        # If it is the last token of the entity,
                        # remove the entity as it is done
                        del entities[0]
                        # Set this so the next token is not removed
                        tkn_part_of_entity = False

                else:
                    if tkn_part_of_entity:
                        del entities[0]
                        tkn_part_of_entity = False

                    if (not ignore_subwords or (
                            self.id2type is not None and
                            self.id2type[tokens['input_ids'][ind]] == 'start'
                            )):
                        labels.append(self.label_map["O"])
                    else:
                        labels.append(self.label_map['X'])

                if len(input_ids) >= self.max_len:
                    logger.debug(
                        "Document exceeding max length encountered. Length of"
                        " current document is %d. Performing chunking...",
                        len(tokens['offset_mapping']))
                    # Split into multiple examples if too long
                    examples['input_ids'].append(input_ids)
                    examples['labels'].append(labels)
                    examples['id'].append(old_ids[_ind])
                    examples['name'].append(old_names[_ind])

                    input_ids = []
                    labels = []

            if input_ids:
                examples['input_ids'].append(input_ids)
                examples['labels'].append(labels)
                examples['id'].append(old_ids[_ind])
                examples['name'].append(old_names[_ind])

        return examples

    def save(self, path: str) -> None:
        with open(path, 'wb') as f:
            dill.dump(self.__dict__, f)

    def ensure_tokenizer(self) -> PreTrainedTokenizerBase:
        if self.hf_tokenizer is None:
            raise ValueError("The tokenizer is not loaded yet")
        return self.hf_tokenizer

    @classmethod
    def load(cls, path: str) -> 'TransformersTokenizer':
        tokenizer = cls()
        with open(path, 'rb') as f:
            d = dill.load(f)
            for k in tokenizer.__dict__:
                if k in d:
                    tokenizer.__dict__[k] = d[k]
        return tokenizer



import logging
import os
from typing import Optional, Union, overload
from transformers.models.bert.tokenization_bert_fast import BertTokenizerFast

from medcat.components.addons.meta_cat.mctokenizers.tokenizers import (
    TokenizerWrapperBase)


FAKE_TOKENIZER_PATH = "#\n/fake-path-not-exist#/"


class TokenizerWrapperBERT(TokenizerWrapperBase):
    """Wrapper around a huggingface BERT tokenizer so that it works with the
    MetaCAT models.

    Args:
        transformers.models.bert.tokenization_bert_fast.BertTokenizerFast:
            A huggingface Fast BERT.
    """
    name = 'bert-tokenizer'

    def __init__(self, hf_tokenizers: Optional[BertTokenizerFast] = None
                 ) -> None:
        super().__init__(hf_tokenizers)

    @overload
    def __call__(self, text: str) -> dict: ...

    @overload
    def __call__(self, text: list[str]) -> list[dict]: ...

    def __call__(self, text: Union[str, list[str]]) -> Union[dict, list[dict]]:
        self.hf_tokenizers = self.ensure_tokenizer()
        if isinstance(text, str):
            result = self.hf_tokenizers.encode_plus(
                text, return_offsets_mapping=True, add_special_tokens=False)

            return {'offset_mapping': result['offset_mapping'],
                    'input_ids': result['input_ids'],
                    'tokens':  self.hf_tokenizers.convert_ids_to_tokens(
                        result['input_ids']),
                    }
        elif isinstance(text, list):
            results = self.hf_tokenizers._batch_encode_plus(
                text, return_offsets_mapping=True, add_special_tokens=False)
            output = []
            for ind in range(len(results['input_ids'])):
                output.append({
                    'offset_mapping': results['offset_mapping'][ind],
                    'input_ids': results['input_ids'][ind],
                    'tokens':  self.hf_tokenizers.convert_ids_to_tokens(
                        results['input_ids'][ind]),
                })
            return output
        else:
            raise Exception("Unsupported input type, supported: text/list, "
                            f"but got: {type(text)}")

    def save(self, dir_path: str) -> None:
        self.hf_tokenizers = self.ensure_tokenizer()
        path = os.path.join(dir_path, self.name)
        self.hf_tokenizers.save_pretrained(path)

    @classmethod
    def load(cls, dir_path: str, model_variant: Optional[str] = '', **kwargs
             ) -> "TokenizerWrapperBERT":
        tokenizer = cls()
        if dir_path != FAKE_TOKENIZER_PATH:
            path = os.path.join(dir_path, cls.name)
            tokenizer.hf_tokenizers = BertTokenizerFast.from_pretrained(
                path, **kwargs)
        else:
            # NOTE: the variable is a string since it's called from meta_cat.py
            #       using a string (in 2 places), but the super class requires
            #       the argument here to be `Optional`.
            variant = str(model_variant)
            logging.warning("Could not load tokenizer (no path provided). "
                            f"Loading from library for model variant: "
                            f"{variant}")
            tokenizer.hf_tokenizers = BertTokenizerFast.from_pretrained(
                variant)
        return tokenizer

    @classmethod
    def create_new(cls, model_variant: Optional[str]
                   ) -> 'TokenizerWrapperBERT':
        return cls.load(FAKE_TOKENIZER_PATH, model_variant)

    def get_size(self) -> int:
        self.hf_tokenizers = self.ensure_tokenizer()
        return len(self.hf_tokenizers.vocab)

    def token_to_id(self, token: str) -> Union[int, list[int]]:
        self.hf_tokenizers = self.ensure_tokenizer()
        return self.hf_tokenizers.convert_tokens_to_ids(token)

    def get_pad_id(self) -> Optional[int]:
        self.hf_tokenizers = self.ensure_tokenizer()
        return self.hf_tokenizers.pad_token_id



from medcat.utils.import_utils import ensure_optional_extras_installed
import medcat


# NOTE: the _ is converted to - automatically
_EXTRA_NAME = "rel-cat"


ensure_optional_extras_installed(medcat.__name__, _EXTRA_NAME)



from medcat.cat import CAT
from medcat.cdb.cdb import CDB
from medcat.preprocessors.cleaners import prepare_name, NameDescriptor

import logging


logger = logging.getLogger(__name__)


def has_per_concept_subnames(cdb: CDB) -> bool:
    for ci in cdb.cui2info.values():
        if ci['subnames']:
            return True
    return False


def _fix_subnames(cat: CAT) -> None:
    tknzer = cat._pipeline.tokenizer_with_tag
    for ci in cat.cdb.cui2info.values():
        names: dict[str, NameDescriptor] = {}
        for name in ci['names']:
            prepare_name(
                name, tknzer, names, (cat.config.general,
                                      cat.config.preprocessing,
                                      cat.config.cdb_maker))
        for name, descr in names.items():
            ci['subnames'].update(descr.snames)
            cat.cdb._subnames.update(descr.snames)


def fix_old_style_cnf(data: dict,
                      remove: set[str] = {"py/object", "__fields_set__",
                                          "__private_attribute_values__"},
                      take_from: str = "py/state.__dict__"):
    all_keys = set(sub_key for key in data for sub_key in
                   (data[key] if isinstance(data[key], dict) else [key]))
    # add 1st level keys
    all_keys.update(data.keys())
    # is old if py/object and py/state somewhere in keys
    if not set(('py/object', 'py/state')) <= all_keys:
        return data
    for to_rm in remove:
        if to_rm in data:
            del data[to_rm]
    # get the data from internal data structure
    cdata = data
    cpath = take_from
    while "." in cpath:
        cur_key, cpath = cpath.split(".", 1)
        if cur_key not in cdata:
            break
        cdata = cdata.pop(cur_key)
    if cpath in cdata:
        cdata = cdata.pop(cpath)
    # do recursive fix and get from internal structure where applicable
    for k, v in cdata.items():
        if isinstance(v, dict):
            v = fix_old_style_cnf(v)
        data[k] = v
    return data


def fix_subnames(cat: CAT) -> None:
    # NOTE: old v1 models may have not stored subnames (snames)
    #       on a per concept basis so we may need to rebuild that
    if has_per_concept_subnames(cat.cdb):
        return
    logger.info(
        "The previous CAT had no per-concept subnames. Adding them now.")
    _fix_subnames(cat)



import json
from typing import Any, cast, Optional, Type
import logging

from pydantic import BaseModel

from medcat.config import Config

from medcat.utils.legacy.helpers import fix_old_style_cnf
from medcat.config.config import SerialisableBaseModel


logger = logging.getLogger(__name__)


SET_IDENTIFIER = "==SET=="
CONFIG_KEEP_IDENTICAL = {
    'cdb_maker', 'preprocessing'
}
CONFIG_MOVE = {
    'linking': 'components.linking',
    'ner': 'components.ner',
    'version.description': 'meta.description',
    'version.id': 'meta.hash',
    'version.ontology': 'meta.ontology',
    'general.spacy_model': 'general.nlp.modelname',
    'general.spacy_disabled_components': 'general.nlp.disabled_components',
}
CONFIG_MOVE_OPTIONAL = {
    "version.description", "version.id", "version.ontology"}
MOVE_WITH_REMOVES = {
    'general': {'checkpoint',  # TODO: Start supporitn checkpoints again
                'spacy_model', 'spacy_disabled_components', 'usage_monitor'},
    'annotation_output': {'doc_extended_info'},
}


def get_val_and_parent_model(old_data: Optional[dict],
                             cnf: Optional[Config],
                             path: str
                             ) -> tuple[Optional[Any], Optional[BaseModel]]:
    """Get the value and the model to set it for from the path specified.

    The paths may be specified in a `.`-separated manner. This unwraps that
    and figures out the value in the old model and the class that should
    be used in the new model.

    Args:
        old_data (Optional[dict]): The raw v1 config data.
        cnf (Optional[Config]): The v2 config.
        path (str): The path to look for.

    Returns:
        tuple[Optional[Any], Optional[BaseModel]]: The value to set, and the
            model to set it for.
    """
    val = old_data
    target_model: Optional[BaseModel] = cnf
    name = path
    while name:
        parts = name.split(".", 1)
        cname = parts[0]
        if len(parts) == 2:
            name = parts[1]
            if target_model is not None:
                target_model = cast(BaseModel, getattr(target_model, cname))
        else:
            name = ''
        if val is not None:
            if path in CONFIG_MOVE_OPTIONAL and cname not in val:
                logger.warning(
                    "Optional path '%s' not found in old config. Ignoring",
                    path)
                val = None
                break
            val = val[cname]
    return val, target_model


def _safe_setattr(target_model: BaseModel, fname: str, val: Any) -> None:
    mval = getattr(target_model, fname)
    if isinstance(mval, BaseModel) and isinstance(val, dict):
        for sname, sval in val.items():
            if not hasattr(mval, sname):
                logger.warning("Trying to set '%s' for '%s' but no such "
                               "attribute", sname, type(mval).__name__)
                continue
            _safe_setattr(mval, sname, sval)
    else:
        setattr(target_model, fname, val)


def _move_identicals(cnf: Config, old_data: dict) -> Config:
    for name in CONFIG_KEEP_IDENTICAL:
        val, target_model = get_val_and_parent_model(old_data, cnf, name)
        val = cast(Any, val)
        target_model = cast(BaseModel, target_model)
        fname = name.split(".")[-1]
        logger.info("Setting %s.%s to %s", type(target_model).__name__, fname,
                    type(val).__name__)
        _safe_setattr(target_model, fname, val)
    return cnf


def _move_partials(cnf: Config, old_data: dict) -> Config:
    for path, to_remove in MOVE_WITH_REMOVES.items():
        val, target_model = get_val_and_parent_model(old_data, cnf, path)
        val = cast(Any, val)
        target_model = cast(BaseModel, target_model)
        val = cast(dict, val).copy()
        for remove in to_remove:
            if remove in val:
                del val[remove]
        fname = path.split(".")[-1]
        logger.info("Setting %s while removing %d", path, len(to_remove))
        _safe_setattr(target_model, fname, val)
    return cnf


def _relocate(cnf: Config, old_data: dict) -> Config:
    for orig_path, new_path in CONFIG_MOVE.items():
        orig_val, _ = get_val_and_parent_model(old_data, cnf=None,
                                               path=orig_path)
        _, target_model = get_val_and_parent_model(None, cnf=cnf,
                                                   path=new_path)
        orig_val = cast(Any, orig_val)
        target_model = cast(BaseModel, target_model)
        fname = new_path.split(".")[-1]
        logger.info("Relocating from %s to %s (%s)", orig_path, new_path,
                    type(orig_val).__name__)
        _safe_setattr(target_model, fname, orig_val)
    return cnf


def _sanitise_sets(old_data: dict) -> dict:
    for k in list(old_data):
        v = old_data[k]
        if isinstance(v, dict) and len(v) == 1 and SET_IDENTIFIER in v:
            logger.info("Moving ")
            old_data[k] = set(v[SET_IDENTIFIER])
        elif isinstance(v, dict):
            # in place anyway
            _sanitise_sets(v)
    return old_data


def _make_changes(cnf: Config, old_data: dict) -> Config:
    old_data = _sanitise_sets(old_data)
    cnf = _move_identicals(cnf, old_data)
    cnf = _move_partials(cnf, old_data)
    cnf = _relocate(cnf, old_data)
    return cnf


def get_config_from_nested_dict(old_data: dict) -> Config:
    """Get the v2 config from v1 json data.

    Args:
        old_data (dict): The json (nested dict) data.

    Returns:
        Config: The v 2 config.
    """
    old_data = fix_old_style_cnf(old_data)
    cnf = Config()
    # v1 models always used spacy
    # but we now default to regex
    cnf.general.nlp.provider = 'spacy'
    cnf = _make_changes(cnf, old_data)
    if cnf.general.nlp.modelname in ('spacy_model', 'en_core_sci_md',
                                     'en_core_sci_lg'):
        logger.info("Fixing spacy model. "
                    "Moving from '%s' to 'en_core_web_md'!",
                    cnf.general.nlp.modelname)
        cnf.general.nlp.modelname = 'en_core_web_md'
    return cnf


def get_config_from_old(path: str) -> Config:
    """Convert the saved v1 config into a v2 Config.

    Args:
        path (str): The v1 config path.

    Returns:
        Config: The v2 config.
    """
    with open(path) as f:
        old_cnf_data = json.load(f)
    return get_config_from_nested_dict(old_cnf_data)


def get_config_from_old_per_cls(
        path: str, cls: Type[SerialisableBaseModel]) -> SerialisableBaseModel:
    """Convert the saved v1 config into a v2 Config for a specific class.

    Args:
        path (str): The v1 config path.
        cls (Type[SerialisableBaseModel]): The class to convert to.

    Returns:
        SerialisableBaseModel: The converted config.
    """
    from medcat.config.config_meta_cat import ConfigMetaCAT
    from medcat.config.config_transformers_ner import ConfigTransformersNER
    from medcat.config.config_rel_cat import ConfigRelCAT
    if cls is Config:
        return get_config_from_old(path)
    elif cls is ConfigMetaCAT:
        from medcat.utils.legacy.convert_meta_cat import (
            load_cnf as load_meta_cat_cnf)
        return load_meta_cat_cnf(path)
    elif cls is ConfigTransformersNER:
        from medcat.utils.legacy.convert_deid import (
            get_cnf as load_deid_cnf)
        return load_deid_cnf(path)
    elif cls is ConfigRelCAT:
        from medcat.utils.legacy.convert_rel_cat import (
            load_cnf as load_rel_cat_cnf)
        return load_rel_cat_cnf(path)
    raise ValueError(f"The config at '{path}' is not a {cls.__name__}!")



import logging
from typing import Iterable, Iterator, Any
from functools import lru_cache
from itertools import product

from pydantic import BaseModel

from medcat.cdb.cdb import CDB
from medcat.cdb.concepts import CUIInfo, NameInfo


logger = logging.getLogger(__name__)


class TranslationLayer:
    """The translation layer for translating:
    - CUIs to names
    - names to CUIs
    - type_ids to CUIs
    - CUIs to chil CUIs

    The idea is to decouple these translations from the CDB instance in
    case something changes there.

    Args:
        cui2info (dict[str, CUIInfo]): The map from CUI to names
        name2info (dict[str, NameInfo]): The map from name to CUIs
        cui2type_ids (dict[str, set[str]]): The map from CUI to type_ids
        cui2children (dict[str, set[str]]): The map from CUI to child CUIs
    """

    def __init__(self, cui2info: dict[str, CUIInfo],
                 name2info: dict[str, NameInfo],
                 cui2children: dict[str, set[str]],
                 separator: str, whitespace: str = ' ') -> None:
        self.cui2info = cui2info
        self.name2info = name2info
        self.separator = separator
        self.whitespace = whitespace
        self.type_id2cuis: dict[str, set[str]] = {}
        for cui, ci in self.cui2info.items():
            type_ids = ci["type_ids"]
            for type_id in type_ids:
                if type_id not in self.type_id2cuis:
                    self.type_id2cuis[type_id] = set()
                self.type_id2cuis[type_id].add(cui)
        self.cui2children = cui2children
        for cui in self.cui2info:
            if cui not in cui2children:
                self.cui2children[cui] = set()

    def get_names_of(self, cui: str, only_prefnames: bool) -> list[str]:
        """Get the preprocessed names of a CUI.

        This method preporcesses the names by replacing the separator
        (generally `~`) with the appropriate whitespace (` `).

        If the concept is not in the underlying CDB, an empty list is returned.

        Args:
            cui (str): The concept in question.
            only_prefnames (bool): Whether to only return a preferred name.

        Returns:
            list[str]: The list of names.
        """
        if cui not in self.cui2info:
            logger.warning(
                "CUI %s Is not defined in CDB / translation layer", cui)
            return []
        if only_prefnames:
            return [self.get_preferred_name(cui).replace(
                self.separator, self.whitespace)]
        return [name.replace(self.separator, self.whitespace)
                # NOTE: sorting the order here in case we're using
                #       edits in which case the order of the names
                #       needs to be the same, otherwise different
                #       edits will be used across runs
                for name in sorted(self.cui2info[cui]["names"])]

    def get_preferred_name(self, cui: str) -> str:
        """Get the preferred name of a concept.

        If no preferred name is found, the random 'first' name is selected.

        Args:
            cui (str): The concept ID.

        Returns:
            str: The preferred name.
        """
        if cui not in self.cui2info:
            logger.warning(
                "CUI %s Is not defined in CDB / translation layer", cui)
            return cui
        pref_name = self.cui2info[cui]["preferred_name"]
        if pref_name is None:
            logger.warning("CUI %s does not have a preferred name. "
                           "Using a random 'first' name of all the names", cui)
            return self.get_first_name(cui)
        return pref_name

    def get_first_name(self, cui: str) -> str:
        """Get the preprocessed (potentially) arbitrarily first name
        of the given concept.

        If the concept does not exist, the CUI itself is returned.

        PS: The "first" name may not be consistent across runs since it
        relies on set order.

        Args:
            cui (str): The concept ID.

        Returns:
            str: The first name.
        """
        for name in self.cui2info[cui]["names"]:
            return name.replace(self.separator, self.whitespace)
        return cui

    def get_direct_children(self, cui: str) -> list[str]:
        """Get the direct children of a concept.

        This means only the children, but not grandchildren.

        If the underlying CDB doesn't list children for this CUI,
        an empty list is returned.

        Args:
            cui (str): The concept in question.

        Returns:
            list[str]: The (potentially empty) list of direct children.
        """
        return list(self.cui2children.get(cui, []))

    @lru_cache(maxsize=10_000)
    def get_direct_parents(self, cui: str) -> list[str]:
        """Get the direct parent(s) of a concept.

        PS: This method can be quite a CPU heavy one since it relies
            on running through all the parent-children relationships
            since the child->parent(s) relationship isn't normally
            kept track of.

        Args:
            cui (str): _description_

        Returns:
            list[str]: _description_
        """
        parents = []
        for pot_parent, children in self.cui2children.items():
            if cui in children:
                parents.append(pot_parent)
        return parents

    def get_children_of(self, found_cuis: Iterable[str],
                        cui: str, depth: int = 1) -> list[str]:
        """Get the children of the specifeid CUI in the
        listed CUIs (if they exist).

        Args:
            found_cuis (Iterable[str]): The list of CUIs to look in
            cui (str): The target parent CUI
            depth (int): The depth to carry out the search for

        Returns:
            list[str]: The list of children found
        """
        if cui not in self.cui2children:
            return []  # no children
        children = self.cui2children[cui]
        found_children = []
        for child in children:
            if child in found_cuis:
                found_children.append(child)
        if depth > 1:
            for child in children:
                found_children.extend(self.get_children_of(
                    found_cuis, child, depth - 1))
        return found_children

    @classmethod
    def from_CDB(cls, cdb: CDB) -> 'TranslationLayer':
        """Construct a TranslationLayer object from a context database (CDB).

        This translation layer will refer to the same dicts that the CDB
        refers to. While there is no obvious reason these should be modified,
        it's something to keep in mind.

        Args:
            cdb (CDB): The CDB

        Returns:
            TranslationLayer: The subsequent TranslationLayer
        """
        if 'pt2ch' not in cdb.addl_info:
            logger.warning(
                "No parent to child information presented so "
                "they cannot be used")
            parent2child = {}
        else:
            parent2child = cdb.addl_info['pt2ch']
        return TranslationLayer(
            cui2info=cdb.cui2info,
            name2info=cdb.name2info,
            cui2children=parent2child,
            separator=cdb.config.general.separator)


class TargetPlaceholder(BaseModel):
    """A class describing the options for a specific placeholder.
    """
    placeholder: str
    target_cuis: list[str]
    onlyprefnames: bool = False


class PhraseChanger(BaseModel):
    """The phrase changer.

    This is class used as a preprocessor for phrases with
    multiple placeholders. It allows swapping in the rest
    of the placeholders while leaving in the one that's
    being tested for.
    """
    preprocess_placeholders: list[tuple[str, str]]

    def __call__(self, phrase: str) -> str:
        for placeholder, replacement in self.preprocess_placeholders:
            phrase = phrase.replace(placeholder, replacement)
        return phrase

    @classmethod
    def empty(cls) -> 'PhraseChanger':
        """Gets the empty phrase changer.

        That is a phrase changer that makes no changes to the phrase.

        Returns:
            PhraseChanger: The empty phrase changer.
        """
        return cls(preprocess_placeholders=[])


class TargetedPhraseChanger(BaseModel):
    """The target phrase changer.

    It includes the phrase changer (for preprocessing) along with
    the relevant concept and the placeholder it will replace.
    """
    changer: PhraseChanger
    placeholder: str
    cui: str
    onlyprefnames: bool


class FinalTarget(BaseModel):
    """The final target.

    This involves the final phrase (which (potentially) has other
    placeholder replaced in it), the placeholder to be replaced,
    and the CUI and specific name being used.
    """
    placeholder: str
    cui: str
    name: str
    final_phrase: str


class OptionSet(BaseModel):
    """The targeting option set.

    This describes all the target placeholders and concepts needed.
    """
    options: list[TargetPlaceholder]
    allow_any_combinations: bool = False

    @classmethod
    def from_dict(cls, section: dict[str, Any]) -> 'OptionSet':
        """Construct a OptionSet instance from a dict.

        The assumed structure is:
        {
            'placeholders': [
                {
                'placeholder': <e.g {DIAGNOSIS}'>,
                'cuis': <the CUI>,
                'prefname-only': 'true'
                }, <potentially more>],
            'any-combination': <True or False>
        }

        The prefname-only key is optional.

        Args:
            section (dict[str, Any]): The dict to parse

        Raises:
            ProblematicOptionSetException:
                If incorrect number of CUIs when not allowing any combination
            ProblematicOptionSetException:
                If placeholders not a list
            ProblematicOptionSetException:
                If multiple placehodlers with same place holder

        Returns:
            OptionSet: The resulting OptionSet
        """
        options: list['TargetPlaceholder'] = []
        allow_any_in = section.get('any-combination', 'false')
        if isinstance(allow_any_in, str):
            allow_any_combinations = allow_any_in.lower() == 'true'
        elif isinstance(allow_any_in, bool):
            allow_any_combinations = allow_any_in
        else:
            raise ProblematicOptionSetException(
                f"Unknown 'any-combination' value: {allow_any_in}")
        if 'placeholders' not in section:
            raise ProblematicOptionSetException(
                "Misconfigured - no placeholders")
        section_placeholders = section['placeholders']
        if not isinstance(section_placeholders, list):
            raise ProblematicOptionSetException(
                "Misconfigured - placehodlers not a list "
                f"({section_placeholders})")
        used_ph = set()
        for part in section_placeholders:
            placeholder = part['placeholder']
            if not isinstance(placeholder, str):
                raise ProblematicOptionSetException(
                    f"Unknown placeholder of type {type(placeholder)}. "
                    "Expected a string. Perhaps you need to surrong the "
                    "placeholder with single quotes (') in the yaml? "
                    f"Received: {placeholder}")
            if placeholder in used_ph:
                raise ProblematicOptionSetException(
                    "Misconfigured - multiple identical placeholders")
            used_ph.add(placeholder)
            target_cuis: list[str] = part['cuis']
            if not isinstance(target_cuis, list):
                raise ProblematicOptionSetException(
                    f"Target CUIs not a list ({type(target_cuis)}): "
                    f"{repr(target_cuis)}")
            if 'prefname-only' in part:
                opn = part['prefname-only']
                if isinstance(opn, bool):
                    onlyprefnames = opn
                else:
                    onlyprefnames = str(opn).lower() == 'true'
            else:
                onlyprefnames = False
            option = TargetPlaceholder(
                placeholder=placeholder, target_cuis=target_cuis,
                onlyprefnames=onlyprefnames)
            options.append(option)
        if not options:
            raise ProblematicOptionSetException(
                "Misconfigured - 0 placeholders found (empty list)")
        if not allow_any_combinations:
            # NOTE: need to have same number of target_cuis
            #       for each placeholder
            # NOTE: there needs to be at least on option / placeholder anyway
            nr_of_cuis = [len(opt.target_cuis) for opt in options]
            if not all(nr == nr_of_cuis[0] for nr in nr_of_cuis):
                raise ProblematicOptionSetException(
                    "Unequal number of cuis when any-combination: false: "
                    f"{nr_of_cuis}. When any-combination: false the number of "
                    "CUIs for each placeholder should be equal.")
        return OptionSet(options=options,
                         allow_any_combinations=allow_any_combinations)

    def to_dict(self) -> dict:
        """Convert the OptionSet to a dict.

        Returns:
            dict: The dict representation
        """
        placeholders = [
            {
                'placeholder': opt.placeholder,
                'cuis': opt.target_cuis,
                'prefname-only': str(opt.onlyprefnames),
            }
            for opt in self.options
        ]
        return {
            'placeholders': placeholders,
            'any-combination': str(self.allow_any_combinations)
        }

    def _get_all_combinations(self, cur_opts: TargetPlaceholder,
                              other_opts: list[TargetPlaceholder],
                              translation: TranslationLayer
                              ) -> Iterator[tuple[PhraseChanger, str]]:
        per_ph_nr_of_opts = [len(opt.target_cuis) for opt in other_opts]
        if self.allow_any_combinations:
            # for each option with N target CUIs use 0, ..., N-1
            for choosers in product(*[range(n) for n in per_ph_nr_of_opts]):
                # NOTE: using the 0th name for target CUI
                placeholders = [
                    (opt.placeholder,
                     translation.get_preferred_name(opt.target_cuis[cui_nr]))
                    for opt, cui_nr in zip(other_opts, choosers)]
                for target_cui in cur_opts.target_cuis:
                    yield (
                        PhraseChanger(preprocess_placeholders=placeholders),
                        target_cui
                    )
        else:
            nr_of_opts = len(cur_opts.target_cuis)
            for cui_nr in range(nr_of_opts):
                placeholders = [
                    # NOTE: using the 0th name for the target CUI
                    (opt.placeholder,
                     translation.get_preferred_name(opt.target_cuis[cui_nr]))
                    for opt in other_opts
                ]
                yield (
                    PhraseChanger(preprocess_placeholders=placeholders),
                    cur_opts.target_cuis[cui_nr]
                )

    def estimate_num_of_subcases(self) -> int:
        """Get the number of distinct subcases.

        This includes ones that can be calculated without the knowledge of the
        underlying CDB. I.e it doesn't care for the number of names involved
        per CUI but only takes into account what is described in the option
        set itself.

        If any combination is allowed, then the answer is the combination of
        the number of target concepts per option. If any combination is not
        allowed, then the answer is simply the number of target concepts for
        an option (they should all have the same number).

        Returns:
            int: Te number of subcases.
        """
        num_of_opts = len(self.options)
        if self.allow_any_combinations:
            total_cases = 1
            for cur_opt in self.options:
                total_cases *= len(cur_opt.target_cuis)
        else:
            total_cases = len(self.options[0].target_cuis)
        return num_of_opts * total_cases

    def get_preprocessors_and_targets(self, translation: TranslationLayer
                                      ) -> Iterator[TargetedPhraseChanger]:
        """Get the targeted phrase changers.

        Args:
            translation (TranslationLayer): The translaton layer.

        Yields:
            Iterator[TargetedPhraseChanger]: Thetarget phrase changers.
        """
        num_of_opts = len(self.options)
        if num_of_opts == 1:
            # NOTE: when there's only 1 option, the other option doesn't work
            #       since it has nothing to iterate over regarding 'other'
            #       options
            opt = self.options[0]
            for target_cui in opt.target_cuis:
                yield TargetedPhraseChanger(changer=PhraseChanger.empty(),
                                            placeholder=opt.placeholder,
                                            cui=target_cui,
                                            onlyprefnames=opt.onlyprefnames)
            return
        for opt_nr in range(num_of_opts):
            other_opts = list(self.options)
            cur_opt = other_opts.pop(opt_nr)
            for changer, target_cui in self._get_all_combinations(
                    cur_opt, other_opts, translation):
                yield TargetedPhraseChanger(
                    changer=changer,
                    placeholder=cur_opt.placeholder,
                    cui=target_cui,
                    onlyprefnames=cur_opt.onlyprefnames)


class ProblematicOptionSetException(ValueError):

    def __init__(self, *args: object) -> None:
        super().__init__(*args)



import os
from transformers import LlamaTokenizerFast
import logging

from medcat.config.config_rel_cat import ConfigRelCAT
from medcat.components.addons.relation_extraction.tokenizer import (
    BaseTokenizerWrapper)

logger = logging.getLogger(__name__)


class TokenizerWrapperLlama(BaseTokenizerWrapper):
    ''' Wrapper around a huggingface Llama tokenizer so that it works with the
    RelCAT models.

    Args:
        hf_tokenizers (`transformers.LlamaTokenizerFast`):
            A huggingface Fast Llama.
    '''
    name = "tokenizer_wrapper_llama_rel"
    pretrained_model_name_or_path = "meta-llama/Llama-3.1-8B"

    @classmethod
    def load(cls, tokenizer_path: str, relcat_config: ConfigRelCAT, **kwargs
             ) -> "TokenizerWrapperLlama":
        tokenizer = cls()
        path = os.path.join(tokenizer_path, cls.name)

        if tokenizer_path:
            tokenizer.hf_tokenizers = LlamaTokenizerFast.from_pretrained(
                path, **kwargs)
        else:
            relcat_config.general.model_name = (
                cls.pretrained_model_name_or_path)
            tokenizer.hf_tokenizers = LlamaTokenizerFast.from_pretrained(
                pretrained_model_name_or_path=relcat_config.general.model_name)
        return tokenizer



import os
import logging
from transformers import PretrainedConfig

from medcat.config.config_rel_cat import ConfigRelCAT


logger = logging.getLogger(__name__)


class RelExtrBaseConfig(PretrainedConfig):
    """ Base class for the RelCAT models
    """
    name = "base-config-relcat"

    def __init__(self, pretrained_model_name_or_path, **kwargs):
        super().__init__(**kwargs)
        self.model_type = "relcat"
        self.pretrained_model_name_or_path = pretrained_model_name_or_path
        self.hf_model_config: PretrainedConfig = kwargs.get(
            "model_config", PretrainedConfig())

    def to_dict(self):
        output = super().to_dict()
        output["model_type"] = self.model_type
        output["pretrained_model_name_or_path"
               ] = self.pretrained_model_name_or_path
        output["model_config"] = self.hf_model_config
        return output

    def save(self, save_path: str):
        self.hf_model_config.to_json_file(
            os.path.join(save_path, "model_config.json"))

    @classmethod
    def load(cls, pretrained_model_name_or_path: str,
             relcat_config: ConfigRelCAT, **kwargs
             ) -> "RelExtrBaseConfig":

        model_config_path = os.path.join(
            pretrained_model_name_or_path, "model_config.json")
        model_config = RelExtrBaseConfig(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            relcat_config=relcat_config, **kwargs)

        if os.path.exists(model_config_path):
            if "modern-bert" in relcat_config.general.tokenizer_name or \
               "modern-bert" in relcat_config.general.model_name:
                from medcat.components.addons.relation_extraction.modernbert.config import RelExtrModernBertConfig  # noqa
                model_config = RelExtrModernBertConfig.load(
                    model_config_path, relcat_config=relcat_config, **kwargs)
            elif "bert" in relcat_config.general.tokenizer_name or \
                    "bert" in relcat_config.general.model_name:
                from medcat.components.addons.relation_extraction.bert.config import RelExtrBertConfig  # noqa
                model_config = RelExtrBertConfig.load(
                    model_config_path, relcat_config=relcat_config, **kwargs)
            elif "llama" in relcat_config.general.tokenizer_name or \
                    "llama" in relcat_config.general.model_name:
                from medcat.components.addons.relation_extraction.llama.config import RelExtrLlamaConfig  # noqa
                model_config = RelExtrLlamaConfig.load(
                    model_config_path, relcat_config=relcat_config, **kwargs)
        else:
            if pretrained_model_name_or_path:
                pretrained_path = pretrained_model_name_or_path
                model_config.hf_model_config = (
                    PretrainedConfig.from_pretrained(
                        pretrained_model_name_or_path=pretrained_path,
                        **kwargs))
            else:
                model_name = relcat_config.general.model_name
                model_config.hf_model_config = (
                    PretrainedConfig.from_pretrained(
                        pretrained_model_name_or_path=model_name, **kwargs))
            logger.info("Loaded config from : " + model_config_path)

        return model_config



import os
import json

from medcat.trainer import Trainer
from medcat.config import Config
from medcat.vocab import Vocab
from medcat.data.mctexport import MedCATTrainerExport

import unittest

import random
import pandas as pd

from .pipeline.test_pipeline import FakeCDB as BFakeCDB
from .utils.legacy.test_convert_config import TESTS_PATH
from .test_cat import TrainedModelTests


class FakeCDB(BFakeCDB):

    def __init__(self, config):
        super().__init__(config)

    def _add_concept(self, *args, **kwargs) -> None:
        pass


class FakeMutEnt:

    def __init__(self, doc: 'FakeMutDoc',
                 start_index: int, end_index: int):
        self.doc = doc
        self.start_index = start_index
        self.end_index = end_index
        self.to_skip = False
        self.base = self
        self.text = doc.text

    @property
    def lower(self) -> str:
        return self.text.lower()


class FakeMutDoc:

    def __init__(self, text: str):
        self.text = text
        self.base = self

    def isupper(self) -> bool:
        return self.text.isupper()

    def get_tokens(self, start_index: int, end_index: int):
        return FakeMutEnt(self, start_index, end_index)

    def __iter__(self):
        yield self.get_tokens(0, 1)


class FakeComponent:
    pass


class FakePipeline:

    def tokenizer(self, text: str) -> FakeMutDoc:
        return FakeMutDoc(text)

    def tokenizer_with_tag(self, text: str) -> FakeMutDoc:
        return FakeMutDoc(text)

    def get_component(self, comp_type):
        return FakeComponent


class TrainerTestsBase(unittest.TestCase):
    DATA_CNT = 14
    TRAIN_DATA = [
        "TEXT#{num}" for num in range(DATA_CNT)
    ]
    DATA_GEN = (dp for dp in TRAIN_DATA)

    @classmethod
    def setUpClass(cls):
        cls.cnf = Config()
        cls.cdb = FakeCDB(cls.cnf)
        cls.vocab = Vocab()
        cls.trainer = Trainer(cls.cdb,
                              cls.caller, FakePipeline())

    def setUp(self):
        self.cnf = Config()
        self.cdb.config = self.cnf
        self.trainer.config = self.cnf

    @classmethod
    def caller(cls, text: str):
        return FakeMutDoc(text)

    @classmethod
    def unlinker(cls, *args, **kwargs):
        return

    @classmethod
    def adder(cls, *args, **kwargs):
        return

    def assert_remembers_training_data(self,
                                       num_docs: int,
                                       num_epochs: int,
                                       unsup: bool = True,
                                       exp_total: int = 1):
        if unsup:
            trained = self.cnf.meta.unsup_trained
        else:
            trained = self.cnf.meta.sup_trained
        self.assertEqual(len(trained), exp_total)
        last_trained = trained[0]
        self.assertEqual(last_trained.num_docs, num_docs)
        self.assertEqual(last_trained.num_epochs, num_epochs)


class TrainerUnsupervisedTests(TrainerTestsBase):
    NEPOCHS = 1
    UNSUP = True

    def train(self, data):
        if self.UNSUP:
            return self.trainer.train_unsupervised(data, nepochs=self.NEPOCHS)
        else:
            return self.trainer.train_supervised_raw(data)

    def test_training_gets_remembered_list(self):
        self.train(self.TRAIN_DATA)
        self.assert_remembers_training_data(self.DATA_CNT, self.NEPOCHS,
                                            unsup=self.UNSUP)

    def test_training_gets_remembered_gen(self):
        self.train(self.DATA_GEN)
        self.assert_remembers_training_data(self.DATA_CNT, self.NEPOCHS,
                                            unsup=self.UNSUP)

    def test_training_gets_remembered_multi(self, repeats: int = 3):
        for _ in range(repeats):
            self.train(self.TRAIN_DATA)
        self.assert_remembers_training_data(self.DATA_CNT, self.NEPOCHS,
                                            exp_total=repeats,
                                            unsup=self.UNSUP)


class TrainerSupervisedTests(TrainerUnsupervisedTests):
    DATA_CNT = 1
    UNSUP = False
    TRAIN_DATA: MedCATTrainerExport = {
        "projects": [
            {
                'cuis': '',
                'tuis': '',
                'documents': [
                    {
                        'id': "P1D1",
                        'name': "Project#1Doc#1",
                        'last_modified': 'N/A',
                        'text': 'Some long text',
                        'annotations': [
                            {
                                'cui': "C1",
                                'start': 0,
                                'end': 4,
                                'value': 'SOME',
                            }
                        ]
                    }
                ],
                'id': "PID#1",
                'name': "PROJECT#1",
            }
        ]
    }

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    def test_training_gets_remembered_gen(self):
        pass  # NOTE: no generation for supervised training


class FromSratchBase(TrainedModelTests):
    RNG_SEED = 42

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.model.cdb.reset_training()
        random.seed(cls.RNG_SEED)


class TrainFromScratchTests(FromSratchBase):
    UNSUP_DATA_PATH = os.path.join(
        TESTS_PATH, "resources", "selfsupervised_data.txt")

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.all_words = list(cls.model.vocab.vocab.keys())
        cls.all_concepts = [(cui, cls.model.cdb.get_name(cui))
                            for cui in cls.model.cdb.cui2info]
        cls.model.trainer.train_unsupervised(cls.get_data())

    @classmethod
    def get_data(cls) -> list[str]:
        df = pd.read_csv(cls.UNSUP_DATA_PATH)
        return df['text'].tolist()

    def test_can_train_unsupervised(self):
        for cui, _ in self.all_concepts:
            with self.subTest(cui):
                self.assertGreater(
                    self.model.cdb.cui2info[cui]['count_train'], 0)


class TrainFromScratchSupervisedTests(TrainFromScratchTests):
    SUP_DATA_PATH = os.path.join(
        TESTS_PATH, "resources", "mct_export_for_test_exp_perfect.json"
    )

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.cnts_before = {
            cui: info['count_train']
            for cui, info in cls.model.cdb.cui2info.items()
        }
        cls.model.trainer.train_supervised_raw(
            cls.get_sup_data()
        )

    @classmethod
    def get_sup_data(cls) -> MedCATTrainerExport:
        with open(cls.SUP_DATA_PATH) as f:
            return json.load(f)

    def test_has_trained_all(self):
        for cui, prev_count in self.cnts_before.items():
            with self.subTest(cui):
                info = self.model.cdb.cui2info[cui]
                self.assertGreater(info['count_train'], prev_count)



from typing import Optional, Protocol, Callable, runtime_checkable, Union
from typing_extensions import Self
from enum import Enum, auto

from medcat.utils.registry import Registry, MedCATRegistryException
from medcat.tokenizing.tokens import MutableDocument, MutableEntity
from medcat.tokenizing.tokenizers import BaseTokenizer
from medcat.cdb import CDB
from medcat.vocab import Vocab
from medcat.config.config import ComponentConfig


class CoreComponentType(Enum):
    tagging = auto()
    token_normalizing = auto()
    ner = auto()
    linking = auto()


@runtime_checkable
class BaseComponent(Protocol):

    @property
    def full_name(self) -> Optional[str]:
        """Name with the component type (e.g ner, linking, meta)."""
        pass

    @property
    def name(self) -> str:
        """The name of the component."""
        pass

    def is_core(self) -> bool:
        """Whether the component is a core component or not.

        Returns:
            bool: Whether this is a core component.
        """
        pass

    def __call__(self, doc: MutableDocument) -> MutableDocument:
        pass

    @classmethod
    def create_new_component(
            cls, cnf: ComponentConfig, tokenizer: BaseTokenizer,
            cdb: CDB, vocab: Vocab, model_load_path: Optional[str]) -> Self:
        """Create a new component or load one off disk if load path presented.

        This may raise an exception if the wrong type of config is provided.

        Args:
            cnf (ComponentConfig): The config relevant to this components.
            tokenizer (BaseTokenizer): The base tokenizer.
            cdb (CDB): The CDB.
            vocab (Vocab): The Vocab.
            model_load_path (Optional[str]): Model load path (if present).

        Returns:
            Self: The new components.
        """
        pass


@runtime_checkable
class CoreComponent(BaseComponent, Protocol):

    def get_type(self) -> CoreComponentType:
        pass


class AbstractCoreComponent(CoreComponent):
    NAME_PREFIX = "core_"

    @property
    def full_name(self) -> str:
        return self.get_type().name + ":" + str(self.name)

    def is_core(self) -> bool:
        return True


@runtime_checkable
class HashableComponet(Protocol):

    def get_hash(self) -> str:
        pass


@runtime_checkable
class TrainableComponent(Protocol):

    def train(self, cui: str,
              entity: MutableEntity,
              doc: MutableDocument,
              negative: bool = False,
              names: Union[list[str], dict] = []) -> None:
        """Train the component.

        This should only apply to the linker.

        Args:
            cui (str): The CUI to train.
            entity (BaseEntity): The entity we're at.
            doc (BaseDocument): The document within which we're working.
            negative (bool): Whether or not the example is negative.
                Defaults to False.
            names (list[str]/dict):
                Optionally used to update the `status` of a name-cui
                pair in the CDB.
        """
        pass


_DEFAULT_TAGGERS: dict[str, tuple[str, str]] = {
    "default": ("medcat.components.tagging.tagger",
                "TagAndSkipTagger.create_new_component"),
}
_DEFAULT_NORMALIZERS: dict[str, tuple[str, str]] = {
    "default": ("medcat.components.normalizing.normalizer",
                "TokenNormalizer.create_new_component"),
}
_DEFAULT_NER: dict[str, tuple[str, str]] = {
    "default": ("medcat.components.ner.vocab_based_ner",
                "NER.create_new_component"),
    "dict": ("medcat.components.ner.dict_based_ner",
             "NER.create_new_component"),
    "transformers_ner": ("medcat.components.ner.trf.transformers_ner",
                         "TransformersNER.create_new_component"),
}
_DEFAULT_LINKING: dict[str, tuple[str, str]] = {
    "default": ("medcat.components.linking.context_based_linker",
                "Linker.create_new_component"),
    "no_action": ("medcat.components.linking.no_action_linker",
                  "NoActionLinker.create_new_component"),
    "medcat2_two_step_linker": (
        "medcat.components.linking.two_step_context_based_linker",
        "TwoStepLinker.create_new_component")
}


_CORE_REGISTRIES: dict[CoreComponentType, Registry[CoreComponent]] = {
    CoreComponentType.tagging: Registry(
        CoreComponent, lazy_defaults=_DEFAULT_TAGGERS),  # type: ignore
    CoreComponentType.token_normalizing: Registry(
        CoreComponent, lazy_defaults=_DEFAULT_NORMALIZERS),  # type: ignore
    CoreComponentType.ner: Registry(CoreComponent,  # type: ignore
                                    lazy_defaults=_DEFAULT_NER),
    CoreComponentType.linking: Registry(CoreComponent,  # type: ignore
                                        lazy_defaults=_DEFAULT_LINKING),
}

CompClass = Callable[[ComponentConfig, BaseTokenizer,
                      CDB, Vocab, Optional[str]], CoreComponent]


def register_core_component(comp_type: CoreComponentType,
                            comp_name: str,
                            comp_clazz: CompClass) -> None:
    """Register a new core component.

    Args:
        comp_type (CoreComponentType): The component type.
        comp_name (str): The component name.
        comp_clazz (ComplClass): The component creator.
    """
    _CORE_REGISTRIES[comp_type].register(comp_name, comp_clazz)


def get_core_registry(comp_type: CoreComponentType) -> Registry[CoreComponent]:
    """Get the registry for a core component type.

    Args:
        comp_type (CoreComponentType): The core component type.

    Returns:
        Registry[CoreComponent]: The corresponding registry.
    """
    return _CORE_REGISTRIES[comp_type]


def get_component_creator(comp_type: CoreComponentType,
                          comp_name: str) -> CompClass:
    """Get the component creator.

    Args:
        comp_type (CoreComponentType): The core component type.
        comp_name (str): The component name.

    Returns:
        Callable[..., CoreComponent]: The creator for the component.
    """
    return get_core_registry(comp_type).get_component(comp_name)


def create_core_component(
        comp_type: CoreComponentType, comp_name: str, cnf: ComponentConfig,
        tokenizer: BaseTokenizer, cdb: CDB, vocab: Vocab,
        model_load_path: Optional[str]) -> CoreComponent:
    """Creat a core component.

    Args:
        comp_type (CoreComponentType): The component type.
        comp_name (str): The name of the component.
        cnf (ComponentConfig): The config to be passed to creator.
        tokenizer (BaseTokenizer): The tokenizer to be passed to creator.
        cdb (CDB): The CDB to be passed to creator.
        vocab (Vocab): The vocab to be passed to the creator.
        model_load_path (Optional[str]): The optional load path to be passed
            to the creators.

    Returns:
        CoreComponent: The resulting / created component.
    """
    try:
        comp_getter = get_core_registry(comp_type).get_component(comp_name)
    except MedCATRegistryException as err:
        raise MedCATRegistryException(f"With comp type '{comp_type}'") from err
    return comp_getter(cnf, tokenizer, cdb, vocab, model_load_path)


def get_registered_components(comp_type: CoreComponentType
                              ) -> list[tuple[str, str]]:
    """Get all registered components (name and class name for each).

    Args:
        comp_type (CoreComponentType): The core component type.

    Returns:
        list[tuple[str, str]]: The name and class name for each
            registered component.
    """
    registry = get_core_registry(comp_type)
    return registry.list_components()



import re
from typing import cast, Optional, Iterator, overload, Union, Any, Type
from collections import defaultdict

from medcat.tokenizing.tokens import (
    BaseToken, BaseEntity, BaseDocument,
    MutableDocument, MutableEntity, MutableToken,
    UnregisteredDataPathException)
from medcat.tokenizing.tokenizers import BaseTokenizer
from medcat.config import Config


class Token:

    def __init__(self, document: 'Document',
                 text: str, _text_with_ws: str,
                 start_index: int, token_index: int,
                 is_punct: bool, to_skip: bool) -> None:
        self._doc = document
        self._text = text
        self._text_with_ws = _text_with_ws
        self._start_index = start_index
        self._token_index = token_index
        self._is_punct = is_punct
        self._to_skip = to_skip
        # defaults
        if self.norm is None:
            # force spacy to init ''
            self.norm = ''

    @property
    def is_punctuation(self) -> bool:
        return self._is_punct

    @is_punctuation.setter
    def is_punctuation(self, new_val: bool) -> None:
        self._is_punct = new_val

    @property
    def to_skip(self) -> bool:
        return self._to_skip

    @to_skip.setter
    def to_skip(self, new_val: bool) -> None:
        self._to_skip = new_val

    @property
    def norm(self) -> str:
        return self.lower

    @norm.setter
    def norm(self, new_val: str) -> None:
        pass  # nothing

    @property
    def base(self) -> BaseToken:
        return cast(BaseToken, self)

    @property
    def text(self) -> str:
        return self._text

    @property
    def text_versions(self) -> list[str]:
        return [self.norm, self.lower]

    @property
    def lower(self) -> str:
        return self._text.lower()

    @property
    def is_stop(self) -> bool:
        return False

    @property
    def is_digit(self) -> bool:
        return self._text.isdigit()

    @property
    def is_upper(self) -> bool:
        return self._text.isupper()

    @property
    def tag(self) -> Optional[str]:
        return None

    @property
    def lemma(self) -> str:
        return self.norm

    @property
    def text_with_ws(self) -> str:
        return self._text_with_ws

    @property
    def char_index(self) -> int:
        return self._start_index

    @property
    def index(self) -> int:
        return self._token_index

    def __str__(self):
        return "RE[T]:" + self.text

    def __repr__(self):
        return repr(str(self))

    def __hash__(self) -> int:
        return hash(self.text)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Token):
            return False
        return (
            self._doc is other._doc and
            self.text == other.text and
            self.index == other.index)


class Entity:
    ENTITY_INFO_PREFIX = "Entity:"
    _addon_extension_paths: set[str] = set()

    def __init__(self, document: 'Document',
                 text: str, start_index: int, end_index: int,
                 start_char_index: int, end_char_index: int) -> None:
        self._doc = document
        self._text = text
        self._start_index = start_index
        self._end_index = end_index
        self._start_char_index = start_char_index
        self._end_char_index = end_char_index
        # defaults
        self.link_candidates: list[str] = []
        self.context_similarity: float = 0.0
        self.confidence: float = 0.0
        self.cui = ''
        self.id = -1  # TODO - what's the default?
        self.detected_name = ''

    @property
    def base(self) -> BaseEntity:
        return cast(BaseEntity, self)

    @property
    def text(self) -> str:
        return self._text

    @property
    def label(self) -> int:
        # NOTE: may not need this?
        return -1

    @property
    def start_index(self) -> int:
        return self._start_index

    @property
    def end_index(self) -> int:
        return self._end_index

    @property
    def start_char_index(self) -> int:
        return self._start_char_index

    @property
    def end_char_index(self) -> int:
        return self._end_char_index

    def set_addon_data(self, path: str, val: Any) -> None:
        # NOTE: doc.get_addon_data will raise if not registered
        doc_dict = self._doc.get_addon_data(f"{self.ENTITY_INFO_PREFIX}{path}")
        doc_dict[(self.start_index, self.end_index)] = val

    def has_addon_data(self, path: str) -> bool:
        return bool(self.get_addon_data(path))

    def get_addon_data(self, path: str) -> Any:
        # NOTE: doc.get_addon_data will raise if not registered
        doc_dict = self._doc.get_addon_data(f"{self.ENTITY_INFO_PREFIX}{path}")
        return doc_dict[(self.start_index, self.end_index)]

    def get_available_addon_paths(self) -> list[str]:
        return [path for path in self._addon_extension_paths
                if self.get_addon_data(path)]

    @classmethod
    def register_addon_path(cls, path: str, def_val: Any = None,
                            force: bool = True) -> None:
        # NOTE: registering for document since that should be constant
        # whereas the entities may be created and recreated
        # it'll map the entity start and end index to the value
        def_val_doc: dict = defaultdict(lambda: def_val)
        Document.register_addon_path(
            f"{cls.ENTITY_INFO_PREFIX}{path}", def_val=def_val_doc,
            force=force)
        cls._addon_extension_paths.add(path)

    def __iter__(self) -> Iterator[MutableToken]:
        for tkn in self._doc._tokens[self.start_index: self.end_index]:
            yield tkn

    def __len__(self) -> int:
        # in terms of tokens
        if self.end_index == self.start_index:
            # distinguish between no tokens and 1 token
            # if there's no tokens, the start and end character matches
            return int(self.end_char_index > self.start_char_index)
        return self.end_index - self.start_index

    def __str__(self):
        return "RE[E]:" + self.text

    def __repr__(self):
        return repr(str(self))


class Document:
    _addon_extension_paths: set[str] = set()

    def __init__(self, text: str, tokens: Optional[list[MutableToken]] = None
                 ) -> None:
        self.text = text
        self._tokens = tokens or []
        self.ner_ents: list[MutableEntity] = []
        self.linked_ents: list[MutableEntity] = []

    @property
    def base(self) -> BaseDocument:
        return cast(BaseDocument, self)

    @overload
    def __getitem__(self, index: int) -> MutableToken:
        pass

    @overload
    def __getitem__(self, index: slice) -> MutableEntity:
        pass

    def __getitem__(self, index: Union[int, slice]
                    ) -> Union[MutableToken, MutableEntity]:
        tokens = self._tokens[index]
        if isinstance(tokens, Token):
            return tokens
        elif isinstance(index, slice):
            token_list = cast(list[MutableToken], tokens)
            return _entity_from_tokens(self, token_list,
                                       index.start, index.stop)
        else:
            raise ValueError(f"Unknown index: {index}")

    def __len__(self) -> int:
        return len(self._tokens)

    def get_tokens(self, start_index: int, end_index: int
                   ) -> list[MutableToken]:
        tkns = []
        for tkn in self:
            if (tkn.base.char_index >= start_index and
                    tkn.base.char_index <= end_index):
                tkns.append(tkn)
        return tkns

    def __iter__(self) -> Iterator[MutableToken]:
        yield from self._tokens

    def isupper(self) -> bool:
        return self.text.isupper()

    def set_addon_data(self, path: str, val: Any) -> None:
        if not hasattr(self.__class__, path):
            raise UnregisteredDataPathException(self.__class__, path)
        setattr(self, path, val)

    def has_addon_data(self, path: str) -> bool:
        return bool(self.get_addon_data(path))

    def get_addon_data(self, path: str) -> Any:
        if not hasattr(self.__class__, path):
            raise UnregisteredDataPathException(self.__class__, path)
        return getattr(self, path)

    def get_available_addon_paths(self) -> list[str]:
        return [path for path in self._addon_extension_paths
                if self.get_addon_data(path)]

    @classmethod
    def register_addon_path(cls, path: str, def_val: Any = None,
                            force: bool = True) -> None:
        setattr(cls, path, def_val)
        cls._addon_extension_paths.add(path)

    def __str__(self):
        return "RE[D]:" + self.text

    def __repr__(self):
        return repr(str(self))


def _entity_from_tokens(doc: Document, tokens: list[MutableToken],
                        token_start: int, token_end: int) -> Entity:
    if any(not isinstance(tkn, Token) for tkn in tokens):
        raise ValueError(
            f"Unknown tokens: {[type(tkn) for tkn in tokens]}")
    rtokens = cast(list[Token], tokens)
    text = doc.text
    if rtokens:
        start_char = rtokens[0].char_index
        # end index should need the length of the last token
        end_char = rtokens[-1].char_index + len(rtokens[-1].text)
        text = text[rtokens[0].char_index: end_char]
    elif doc._tokens:
        if token_start >= len(doc._tokens):
            start_char = len(doc.text)
        else:
            start_char = doc._tokens[token_start].base.char_index
        end_char = start_char
        text = ''
    else:
        start_char = end_char = 0
        text = ''
    return Entity(doc, text, token_start, token_end, start_char, end_char)


class RegexTokenizer(BaseTokenizer):
    REGEX = r'((\b\w+\b|\S+)\s?)'
    # group 1: text with whitespace (if present)
    # group 2: text with no whitespace

    def create_entity(self, doc: MutableDocument,
                      token_start_index: int, token_end_index: int,
                      label: str) -> MutableEntity:
        rdoc = cast(Document, doc)
        return self.entity_from_tokens(
            # rdoc._tokens[token_start_index: token_end_index + 1])
            rdoc._tokens[token_start_index: token_end_index])
        # spacy_doc = cast(Document, doc)._delegate
        # span = Span(spacy_doc, token_start_index, token_end_index, label)
        # return Entity(span)

    def entity_from_tokens(self, tokens: list[MutableToken]) -> MutableEntity:
        if not tokens:
            raise ValueError("Need at least one token for an entity")
        doc = cast(Token, tokens[0])._doc
        start_index = doc._tokens.index(tokens[0])
        end_index = doc._tokens.index(tokens[-1])
        return _entity_from_tokens(doc, tokens, start_index, end_index)

    def __call__(self, text: str) -> MutableDocument:
        tokens = re.finditer(self.REGEX, text)
        doc = Document(text)
        for tkn_index, match in enumerate(tokens):
            start_index = match.start()
            token_w_ws = match.group(1)
            token = match.group(2)
            doc._tokens.append(Token(doc, token, token_w_ws,
                                     start_index, tkn_index,
                                     False, False))
        return doc

    @classmethod
    def create_new_tokenizer(cls, config: Config) -> 'RegexTokenizer':
        return cls()

    def get_doc_class(self) -> Type[MutableDocument]:
        return Document

    def get_entity_class(self) -> Type[MutableEntity]:
        return Entity



from typing import Protocol, Any, TypeVar, Generic, cast
from datetime import datetime
from dataclasses import is_dataclass, asdict, dataclass
import importlib
import json

import numpy as np

from medcat.storage.serialisers import Serialiser, AvailableSerialisers


T = TypeVar("T")


# Protocol for type handlers
class TypeHandler(Protocol, Generic[T]):
    type_name: str
    type_cls: type

    def should_encode(self, obj: Any) -> bool:
        pass

    def encode(self, obj: T) -> Any:
        """Encode an object of the registered type."""
        pass

    def decode(self, obj: Any) -> T:
        """Decode an object of the registered type."""
        pass


# Registry for type handlers
class TypeRegistry:
    _type_key = "__type__"
    _data_key = "data"

    def __init__(self) -> None:
        self.handlers: dict[str, TypeHandler] = {}

    def register(self, handler: TypeHandler) -> None:
        """Register a new type handler."""
        self.handlers[handler.type_name] = handler

    def encode(self, obj: Any) -> Any:
        """Encode an object using the registered handler."""
        for handler in self.handlers.values():
            if handler.should_encode(obj):
                return {self._type_key: handler.type_name,
                        self._data_key: handler.encode(obj)}
        return obj  # Default handling

    def decode(self, obj: Any) -> Any:
        """Decode an object using the registered handler."""
        if isinstance(obj, dict) and self._type_key in obj:
            type_name = obj[self._type_key]
            if type_name in self.handlers:
                return self.handlers[type_name].decode(obj[self._data_key])
        return obj  # Default handling


class TypeBasedHandler(TypeHandler[T]):

    def should_encode(self, obj: Any) -> bool:
        return isinstance(obj, self.type_cls)


class NumpyArrayHandler(TypeBasedHandler[np.ndarray]):
    type_name = "ndarray"
    type_cls = np.ndarray
    _dtype_key = "dtype"
    _data_key = "data"
    _shape_key = "shape"

    def encode(self, obj: np.ndarray) -> Any:
        """Encode numpy ndarray."""
        return {
            self._data_key: obj.tolist(),
            self._dtype_key: str(obj.dtype),
            self._shape_key: obj.shape}

    def decode(self, obj: Any) -> np.ndarray:
        """Decode to numpy ndarray."""
        return np.array(
            obj[self._data_key],
            dtype=obj[self._dtype_key]).reshape(
                obj[self._shape_key])


class SetHandler(TypeBasedHandler[set]):
    type_name = "set"
    type_cls = set

    def encode(self, obj: set) -> Any:
        """Encode set."""
        return list(obj)

    def decode(self, obj: Any) -> set:
        """Decode to set."""
        return set(obj)


class DateTimeHandler(TypeBasedHandler[datetime]):
    type_name = "datetime"
    type_cls = datetime

    def encode(self, obj: datetime) -> Any:
        return obj.isoformat()

    def decode(self, obj: Any):
        return datetime.fromisoformat(obj)


class DataClassHandler(TypeHandler[T]):
    type_name = "dataclass"
    type_cls = type(dataclass)  # NOTE: shouldn't be used
    _cls_key = "class-path"
    _data_key = "data"

    def should_encode(self, obj: Any) -> bool:
        return is_dataclass(obj)

    def encode(self, obj: T) -> Any:
        cls_path = f"{obj.__class__.__module__}.{obj.__class__.__name__}"
        return {
            self._cls_key: cls_path,
            # NOTE: mypy doesn't know that this is a dataclass
            self._data_key: asdict(obj),  # type: ignore
        }

    def decode(self, obj: Any) -> T:
        cls_path: str = obj[self._cls_key]
        module_name, cls_name = cls_path.rsplit('.', 1)
        module = importlib.import_module(module_name)
        cls = getattr(module, cls_name)
        if not is_dataclass(cls):
            raise TypeError(f"Unsupported type: {cls}")
        # NOTE: mypy doesn't know that the dataclass init can be called here
        return cast(T, cls(**obj[self._data_key]))  # type: ignore


_def_registry = TypeRegistry()


_def_registry.register(NumpyArrayHandler())
_def_registry.register(SetHandler())
_def_registry.register(DateTimeHandler())
_def_registry.register(DataClassHandler())


class JsonSerialiser(Serialiser):
    ser_type = AvailableSerialisers.json

    def serialise(self, raw_parts: dict[str, Any], target_file: str) -> None:
        with open(target_file, 'w') as f:
            json.dump(raw_parts, f, default=_def_registry.encode)

    def deserialise(self, target_file: str) -> dict[str, Any]:
        with open(target_file, 'r') as f:
            return json.load(f, object_hook=_def_registry.decode)



import logging
import contextlib
from typing import TypedDict, cast
import tempfile
import dill

from copy import deepcopy

from medcat.cdb.concepts import NameInfo, CUIInfo
from medcat.config.config import ModelMeta


logger = logging.getLogger(__name__)


CDBState = TypedDict(
    'CDBState',
    {
        'name2info': dict[str, NameInfo],
        'cui2info': dict[str, CUIInfo],
        'token_counts': dict[str, int],
        '_subnames': set[str],
        'config.meta': ModelMeta,
    })
"""CDB State.

This is a dictionary of the parts of the CDB that change during
(supervised) training. It can be used to store and restore the
state of a CDB after modifying it.

Currently, the following fields are saved:
 - name2info
 - cui2info
 - token_counts
 - _subnames
 - config.meta
"""


def _get_attr(cdb, path: str) -> object:
    cur_obj = cdb
    cur_path = path
    while "." in cur_path:
        cur_left, cur_path = cur_path.split(".", 1)
        cur_obj = getattr(cur_obj, cur_left)
    return getattr(cur_obj, cur_path)


def _set_attr(cdb, path: str, val: object) -> None:
    cur_obj = cdb
    cur_path = path
    while "." in cur_path:
        cur_left, cur_path = cur_path.split(".", 1)
        cur_obj = getattr(cur_obj, cur_left)
    setattr(cur_obj, cur_path, val)


def copy_cdb_state(cdb) -> CDBState:
    """Creates a (deep) copy of the CDB state.

    Grabs the fields that correspond to the state,
    creates deep copies, and returns the copies.

    Args:
        cdb: The CDB from which to grab the state.

    Returns:
        CDBState: The copied state.
    """
    return cast(CDBState, {
        k: deepcopy(_get_attr(cdb, k)) for k in CDBState.__annotations__
    })


def save_cdb_state(cdb, file_path: str) -> None:
    """Saves CDB state in a file.

    Currently uses `dill.dump` to save the relevant fields/values.

    Args:
        cdb: The CDB from which to grab the state.
        file_path (str): The file to dump the state.
    """
    # NOTE: The difference is that we don't create a copy here.
    #       That is so that we don't have to occupy the memory for
    #       both copies
    the_dict = {
        k: _get_attr(cdb, k) for k in CDBState.__annotations__
    }
    logger.debug("Saving CDB state on disk at: '%s'", file_path)
    with open(file_path, 'wb') as f:
        dill.dump(the_dict, f)


def apply_cdb_state(cdb, state: CDBState) -> None:
    """Apply the specified state to the specified CDB.

    This overwrites the current state of the CDB with one provided.

    Args:
        cdb: The CDB to apply the state to.
        state (CDBState): The state to use.
    """
    _clear_state(cdb)
    _reapply_state(cdb, state)


def _clear_state(cdb) -> None:
    for k in CDBState.__annotations__:
        val = _get_attr(cdb, k)
        if not isinstance(val, (dict, set, ModelMeta)):
            raise ValueError(
                "A part of the CDB state was not a dict, set, or ModelMeta "
                f"(during clearing). Got {type(val).__name__}. The "
                "re-setting of the sate needs to be implemented per type.")
        if isinstance(val, (dict, set)):
            val.clear()
        else:
            val.sup_trained.clear()
            val.unsup_trained.clear()


def _reapply_state(cdb, state: CDBState):
    for k, v in state.items():
        # trying to preserve the instances
        prev_ver = _get_attr(cdb, k)
        if (not isinstance(prev_ver, (dict, set, ModelMeta)) or
                not isinstance(v, (dict, set, ModelMeta))):
            raise ValueError(
                "A part of the CDB state was not a dict, set, ModelMeta "
                f"(during setting). Got {type(prev_ver).__name__} | "
                f"{type(v).__name__}. The re-setting of the sate needs to be"
                "implemented per type.")
        if isinstance(prev_ver, (dict, set)):
            prev_ver.update(v)
        elif isinstance(prev_ver, ModelMeta):
            # just set, shouldn't matter
            _set_attr(cdb, k, v)


def load_and_apply_cdb_state(cdb, file_path: str) -> None:
    """Delete current CDB state and apply CDB state from file.

    This first deletes the current state of the CDB.
    This is to save memory. The idea is that saving the staet
    on disk will save on RAM usage. But it wouldn't really
    work too well if upon load, two instances were still in
    memory.

    Args:
        cdb: The CDB to apply the state to.
        file_path (str): The file where the state has been saved to.
    """
    # clear existing data on CDB
    # this is so that we don't occupy the memory for both the loaded
    # and the on-CDB data
    logger.debug("Clearing CDB state in memory")
    _clear_state(cdb)
    logger.debug("Loading CDB state from disk from '%s'", file_path)
    with open(file_path, 'rb') as f:
        state: CDBState = dill.load(f)
    _reapply_state(cdb, state)


@contextlib.contextmanager
def captured_state_cdb(cdb, save_state_to_disk: bool = False):
    """A context manager that captures and re-applies the initial CDB state.

    The context manager captures/copies the initial state of the CDB when
    entering. It then allows the user to modify the state (i.e training).
    Upon exit re-applies the initial CDB state.

    If RAM is an issue, it is recommended to use `save_state_to_disk`.
    Otherwise the copy of the original state will be held in memory.
    If saved on disk, a temporary file is used and removed afterwards.

    Args:
        cdb: The CDB to use.
        save_state_to_disk (bool): Whether to save state on disk or hold
            in memory. Defaults to False.

    Yields:
        None
    """
    if save_state_to_disk:
        with on_disk_memory_capture(cdb):
            yield
    else:
        with in_memory_state_capture(cdb):
            yield


@contextlib.contextmanager
def in_memory_state_capture(cdb):
    """Capture the CDB state in memory.

    Args:
        cdb: The CDB to use.

    Yields:
        None
    """
    state = copy_cdb_state(cdb)
    yield
    apply_cdb_state(cdb, state)


@contextlib.contextmanager
def on_disk_memory_capture(cdb):
    """Capture the CDB state in a temporary file.

    Args:
        cdb: The CDB to use

    Yields:
        None
    """
    with tempfile.NamedTemporaryFile() as tf:
        save_cdb_state(cdb, tf.name)
        yield
        load_and_apply_cdb_state(cdb, tf.name)
